"""
This is the main start up file of the project planner
"""

import argparse
import codecs
import locale
import logging
import sys
from datetime import datetime
from pathlib import Path

import dateutil.parser as dparse
import yaml

from gantt_project_maker import __version__
from gantt_project_maker.colors import set_custom_colors
from gantt_project_maker.project_classes import (
    ProjectPlanner,
    SCALES,
    parse_date,
    extend_suffix,
)

__author__ = "Eelco van Vliet"
__copyright__ = "Eelco van Vliet"
__license__ = "MIT"

_logger = logging.getLogger(__name__)


############################################################################


def get_info_from_file_or_settings(settings, key):
    """
    Get the information directly from the settings or from a separate file if a filename is given
    Parameters
    ----------
    settings: dict
        Settings file
    key: str
        Key of the settings file we want to read

    Returns
    -------
    dict:
        Structure with information from the settings or separate file

    """
    information = settings[key]

    if isinstance(information, str):
        with codecs.open(information, "r", encoding="UTF-8") as stream:
            information = yaml.load(stream=stream, Loader=yaml.Loader)

    return information


def get_pasted_employees(args_employee, employees_info):
    """get the list of full names or requested employees and return as a comma separated string"""
    all_employees = []
    for employee in args_employee:
        employee_name = get_employee_name(employees_info, employee)
        all_employees.append(employee_name)
    all_employees = ", ".join(all_employees)
    return all_employees


def get_employee_name(employees_info, employee):
    """get the full name of the employee from the settings file"""
    try:
        request_employee = employees_info[employee]
    except KeyError as err:
        _logger.warning(err)
        raise KeyError(
            f"Employee {employee} given via argument is not find in section 'employees' "
            f"in settings file. Please add this employee to your settings"
        )
    else:
        try:
            request_name = request_employee["name"]
        except KeyError as err:
            _logger.warning(err)
            raise KeyError("'name' not given for employee {request_employee} ")

    return request_name


def check_if_date(value):
    """check if an argument is a valid date. Return the original string value"""
    try:
        date = dparse.parse(value).date()
    except ValueError:
        raise argparse.ArgumentTypeError(f"Date {value} is not a valid date")
    return value


def parse_args(args):
    """Parse command line parameters

    Args:
      args (List[str]): command line parameters as list of strings
          (for example  ``["--help"]``).

    Returns:
      :obj:`argparse.Namespace`: command line parameters namespace
    """

    parser = argparse.ArgumentParser(
        description="A front end to the python-gantt project planning"
    )
    parser.add_argument("settings_filename", help="Name of the configuration file")
    parser.add_argument("--output_filename", help="Name of the text output file")
    parser.add_argument(
        "--version",
        action="version",
        version=f"gantt_project_maker {__version__}",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="loglevel",
        help="set loglevel to INFO",
        action="store_const",
        const=logging.INFO,
        default=logging.INFO,
    )
    parser.add_argument(
        "-vv",
        "--debug",
        dest="loglevel",
        help="set loglevel to DEBUG",
        action="store_const",
        const=logging.DEBUG,
    )
    parser.add_argument(
        "-q",
        "--quiet",
        dest="loglevel",
        help="set loglevel to WARNING",
        action="store_const",
        const=logging.WARNING,
    )
    parser.add_argument(
        "-vvv",
        "--very_verbose",
        help="Also show the logging of the gantt module",
        action="store_true",
    )
    parser.add_argument(
        "-s",
        "--scale",
        help="The scale of the grid of the project scheme",
        choices=set(SCALES.keys()),
    )
    parser.add_argument(
        "--details",
        help="Add all the tasks with the detail attribute",
        action="store_true",
        default=True,
    )
    parser.add_argument(
        "--no_details",
        help="Suppress all the tasks with the detail attribute.",
        action="store_false",
        dest="details",
    )
    parser.add_argument(
        "-e",
        "--export_to_xlsx",
        help="Export the project plan to Excel",
        action="store_true",
    )
    parser.add_argument(
        "-g",
        "--collaps_tasks",
        help="Collaps the tasks per project to one task to remove details",
        action="store_true",
    )
    parser.add_argument(
        "-b",
        "--resources",
        help="Write the resources file of the planning",
        action="store_true",
    )
    parser.add_argument(
        "--vacations",
        help="Write the vacations file of all the  employees",
        action="store_true",
    )
    parser.add_argument(
        "-m",
        "--employee",
        help="Only use the projects of this employee. Can be given multiple times for multiple "
        "employees.",
        action="append",
    )
    parser.add_argument(
        "-f",
        "--filter_employees",
        help="Only include tasks to which this employee contributes. Can be given multiple times for multiple "
        "employees.",
        action="append",
    )
    parser.add_argument(
        "-p",
        "--period",
        help="On export this period from the list of periods as given in the settings file. If "
        "not given, all the periods are writen to file",
        action="append",
    )
    parser.add_argument(
        "--start_planning",
        type=check_if_date,
        help="Start of the planning. If not given, the value given in de settings file is taken",
    )
    parser.add_argument(
        "--end_planning",
        type=check_if_date,
        help="End of the planning. If not given, the value given in de settings file is taken",
    )
    parser.add_argument(
        "--weeks_margin_left",
        type=int,
        help="Shifts start of planning with this number of weeks left without adding projects. Default is read "
        "from the settings file. This value overrides the default",
    )
    parser.add_argument(
        "--weeks_margin_right",
        type=int,
        help="Shifts start of planning with this number of weeks right without adding projects. Default is read "
        "from the settings file. This value overrides the default.",
    )
    parser.add_argument("--pdf", help="Save the svg also as pdf ", action="store_true")

    return parser.parse_args(args)


def setup_logging(loglevel):
    """Setup basic logging

    Args:
      loglevel (int): minimum loglevel for emitting messages
    """
    if loglevel == logging.DEBUG:
        log_format = "[%(levelname)5s]:%(filename)s/%(lineno)d: %(message)s"
    else:
        log_format = "[%(levelname)s] %(message)s"
    logging.basicConfig(
        level=loglevel,
        stream=sys.stdout,
        format=log_format,
        datefmt="%Y-%m-%d %H:%M:%S",
    )


def check_if_items_are_available(requested_items, available_items, label=""):
    """
    Check is the passed items in the list are available in the keys of the dictionary

    Parameters
    ----------
    requested_items: list
        All requested items  in the list
    available_items: dict
        The dictionary with the keys
    label: str
        Used for information to the screen

    """
    unique_available_items = set(list(available_items.keys()))
    if missing_items := set(requested_items).difference(unique_available_items):
        raise ValueError(
            f"The {label} {missing_items} are not defined in the settings file.\n"
            f"The following keys are available: {unique_available_items}"
        )
    return True


def make_banner(width=80) -> None:
    """
    Make a banner with the start time
    Args:
        width:  int
            With of the banner

    Returns:
    """
    print("-" * 80)
    exe = Path(sys.argv[0]).stem
    now = datetime.now()
    print(
        f"Start '{exe} {' '.join(sys.argv[1:])}'\nat {now.date()} {now.time().strftime('%H:%M')} "
    )
    print("-" * 80)


def main(args):
    args = parse_args(args)

    if args.loglevel < logging.WARNING:
        make_banner()

    setup_logging(args.loglevel)
    if args.very_verbose:
        gantt_logger = logging.getLogger("Gantt")
        gantt_logger.setLevel(args.loglevel)

    _logger.info("Reading settings file {}".format(args.settings_filename))
    with codecs.open(args.settings_filename, "r", encoding="UTF-8") as stream:
        settings = yaml.load(stream=stream, Loader=yaml.Loader)

    general_settings = settings["general"]
    try:
        project_settings_per_project_leader = settings[
            "project_settings_file_per_employee"
        ]
    except KeyError as err:
        _logger.warning(err)
        raise KeyError(
            "Entry project_settings_file_per_employee not found. Are you sure this"
            "is the main settingsfile and not the settings file of an employee?"
        )
    period_info = settings["periods"]
    dayfirst = general_settings["dayfirst"]

    if args.scale is not None:
        scale_key = args.scale
    else:
        scale_key = general_settings.get("scale", "daily")
    scale = SCALES[scale_key]

    if args.start_planning is None:
        start = parse_date(general_settings["planning_start"], dayfirst=dayfirst)
    else:
        start = parse_date(args.start_planning, dayfirst=dayfirst)
    if args.end_planning is None:
        end = parse_date(general_settings["planning_end"], dayfirst=dayfirst)
    else:
        end = parse_date(args.end_planning, dayfirst=dayfirst)

    if args.weeks_margin_left is None:
        weeks_margin_left = general_settings.get("weeks_margin_left")
    else:
        weeks_margin_left = args.weeks_margin_left

    if args.weeks_margin_right is None:
        weeks_margin_right = general_settings.get("weeks_margin_right")
    else:
        weeks_margin_right = args.weeks_margin_right

    programma_title = general_settings["title"]
    programma_color = general_settings.get("color")
    vacation_color = general_settings.get("vacation_color", programma_color)
    output_directories = general_settings.get("output_directories")
    excel_info = settings.get("excel")
    employees_info = get_info_from_file_or_settings(settings=settings, key="employees")
    vacations_info = get_info_from_file_or_settings(settings=settings, key="vacations")

    if args.employee:
        all_employees = get_pasted_employees(
            args.employee, employees_info=employees_info
        )
        programma_title += f"/{all_employees}"

    if args.filter_employees:
        all_filter_employees = get_pasted_employees(
            args.filter_employees, employees_info=employees_info
        )
        programma_title += f": {all_filter_employees}"

    fill = "black"
    stroke = "black"
    stroke_width = 0
    font_family = "Verdana"
    if font_info := general_settings.get("font_info"):
        fill = font_info.get("fill", fill)
        stroke = font_info.get("stroke", stroke)
        stroke_width = font_info.get("stroke_width", stroke_width)
        font_family = font_info.get("font_family", font_family)

    if custom_colors := general_settings.get("custom_colors"):
        set_custom_colors(custom_colors=custom_colors)

    vacations_title_default = "Vacations"
    if country_code := general_settings.get("country_code"):
        locale.setlocale(locale.LC_TIME, country_code)
        if country_code.startswith("nl"):
            vacations_title_default = "Vakanties"
    vacations_title = general_settings.get("vacations_title", vacations_title_default)

    if output_directories is not None:
        planning_directory = Path(output_directories.get("planning", "."))
        resources_directory = Path(output_directories.get("resources", "."))
        excel_directory = Path(output_directories.get("excel", "."))
        vacations_directory = Path(output_directories.get("vacations", "."))
    else:
        planning_directory = Path(".")
        resources_directory = Path(".")
        excel_directory = Path(".")
        vacations_directory = Path(".")

    if args.employee is not None:
        check_if_items_are_available(
            requested_items=args.employee,
            available_items=project_settings_per_project_leader,
            label="employee project",
        )
    if args.filter_employees is not None:
        check_if_items_are_available(
            requested_items=args.filter_employees,
            available_items=employees_info,
            label="employee task",
        )

    if args.period is not None:
        check_if_items_are_available(
            requested_items=args.period, available_items=period_info, label="period"
        )

    # read the settings file per employee
    settings_per_project_leader = {}
    for (
        project_leader_key,
        employee_settings_file,
    ) in project_settings_per_project_leader.items():
        _logger.info(
            f"Reading settings file {employee_settings_file} of  employee {project_leader_key}"
        )
        with codecs.open(employee_settings_file, "r", encoding="UTF-8") as stream:
            settings_per_project_leader[project_leader_key] = yaml.load(
                stream=stream, Loader=yaml.Loader
            )

    if args.output_filename is None:
        output_filename = Path(args.settings_filename).with_suffix(".svg")
    else:
        output_filename = Path(args.output_filename).with_suffix(".svg")

    if args.employee is not None:
        output_filename = extend_suffix(
            output_filename=output_filename, extensions=args.employee
        )

    if args.filter_employees is not None:
        extensions = ["contributors"] + sorted(args.filter_employees)
        output_filename = extend_suffix(
            output_filename=output_filename, extensions=extensions
        )

    today = None
    try:
        today_reference = general_settings["reference_date"]
    except KeyError:
        _logger.debug("No date found")
    else:
        if today_reference is not None:
            if today_reference == "today":
                today = datetime.today().date()
                _logger.debug("Setting date to today {}".format(today))
            else:
                today = parse_date(today_reference, dayfirst=dayfirst)
                _logger.debug("Setting date to {}".format(today))
        else:
            _logger.debug("today key found be no date defined")

    # Begin de planning
    planning = ProjectPlanner(
        programma_title=programma_title,
        vacations_title=vacations_title,
        programma_color=programma_color,
        vacation_color=vacation_color,
        output_file_name=output_filename,
        planning_start=start,
        planning_end=end,
        weeks_margin_left=weeks_margin_left,
        weeks_margin_right=weeks_margin_right,
        today=today,
        dayfirst=dayfirst,
        scale=scale,
        period_info=period_info,
        excel_info=excel_info,
        details=args.details,
        filter_employees=args.filter_employees,
        save_svg_as_pdf=args.pdf,
        collaps_tasks=args.collaps_tasks,
    )

    # add global information, vacations and employees
    planning.add_global_information(
        fill=fill, stroke=stroke, stroke_width=stroke_width, font_family=font_family
    )
    planning.add_vacations(vacations_info=vacations_info)
    planning.add_employees(employees_info=employees_info)

    # Add the general tasks per employee. It is not mandatory to add tasks_and_milestones,
    # however, you may. The advantage is that multiply tasks can share the same milestone
    for (
        project_leader_key,
        project_leader_settings,
    ) in settings_per_project_leader.items():
        if tasks_and_milestones_info := project_leader_settings.get(
            "tasks_and_milestones"
        ):
            _logger.info(f"Adding global tasks en milestones of {project_leader_key} ")
            planning.add_tasks_and_milestones(
                tasks_and_milestones_info=tasks_and_milestones_info
            )

    # Voeg nu de projecten per employee toe.
    for (
        project_leader_key,
        project_leader_settings,
    ) in settings_per_project_leader.items():
        if args.employee is not None and project_leader_key not in args.employee:
            _logger.debug(f"Skip employee {project_leader_key}")
            continue

        project_employee_info = project_leader_settings["general"]
        subprojects_info = project_leader_settings["projects"]

        subprojects_selection = project_employee_info["projects"]
        subprojects_title = project_employee_info["title"]
        subprojects_color = project_employee_info.get("color")
        planning.make_projects(
            subprojects_info=subprojects_info,
            subprojects_selection=subprojects_selection,
            subprojects_title=subprojects_title,
            subprojects_color=subprojects_color,
        )

    # Everything has been added to the planning. Write it to file
    planning.write_planning(
        write_resources=args.resources,
        write_vacations=args.vacations,
        planning_output_directory=planning_directory,
        resource_output_directory=resources_directory,
        vacations_output_directory=vacations_directory,
        periods=args.period,
    )

    if args.export_to_xlsx:
        planning.exporteer_naar_excel(excel_output_directory=excel_directory)


def run():
    """Calls :func:`main` passing the CLI arguments extracted from :obj:`sys.argv`

    This function can be used as entry point to create console scripts with setuptools.
    """
    main(sys.argv[1:])


if __name__ == "__main__":
    # ^  This is a guard statement that will prevent the following code from
    #    being executed in the case someone imports this file instead of
    #    executing it as a script.
    #    https://docs.python.org/3/library/__main__.html

    # After installing your project with pip, users can also run your Python
    # modules as scripts via the ``-m`` flag, as defined in PEP 338::
    #
    #     python -m gantt_projectplanner.skeleton 42
    #
    run()



import pytest

from gantt_project_maker.colors import hex_number_to_hex_hash, color_to_hex

__author__ = "Eelco van Vliet"
__copyright__ = "Eelco van Vliet"
__license__ = "MIT"


def test_hex_number_to_hex_hash():
    assert hex_number_to_hex_hash("AABBCC") == "#AABBCC"
    assert hex_number_to_hex_hash("#AABBCC") == "#AABBCC"


def test_hex_number_to_hex_hash_error_wrong_char():
    with pytest.raises(ValueError):
        hex_number_to_hex_hash("AABBCX")


def test_hex_number_to_hex_hash_error_wrong_place():
    with pytest.raises(ValueError):
        hex_number_to_hex_hash("AA#BCC")


def test_hex_number_to_hex_hash_error_wrong_number():
    with pytest.raises(ValueError):
        hex_number_to_hex_hash("AABBCCE")


def test_color_to_hex():
    assert color_to_hex("black") == "#000000"
    assert color_to_hex("navy") == "#000080"
    assert color_to_hex("pink").upper() == "#FFC0CB"



import logging

import pandas as pd
from pandas.io.formats.excel import ExcelFormatter

import gantt_project_maker.gantt as gantt
from gantt_project_maker.colors import color_to_hex

_logger = logging.getLogger(__name__)

PAGE_WIDTH = 100
CHAR_PER_LINE = 112


class WorkBook:
    def __init__(self, workbook):
        self.workbook = workbook
        self.left_align_italic = None
        self.left_align_italic_large = None
        self.left_align_italic_large_ul = None
        self.left_align_helvetica = None
        self.left_align_helvetica_bold = None
        self.left_align_bold = None
        self.left_align_bold_large = None
        self.left_align_bold_larger = None
        self.left_align = None
        self.left_align_large_wrap = None
        self.left_align_large_wrap_top = None
        self.left_align_wrap = None
        self.left_align_large = None
        self.right_align = None
        self.header_format = None
        self.title_format = None
        self.section_heading = None
        self.footer_format = None
        self.merge_format = None
        self.date_format = None
        self.add_styles()

    def add_styles(self):
        self.left_align_helvetica = self.workbook.add_format(
            {"font": "helvetica", "align": "left", "font_size": 8, "border": 0}
        )
        self.left_align_helvetica_bold = self.workbook.add_format(
            {
                "font": "helvetica",
                "bold": True,
                "align": "left",
                "font_size": 8,
                "border": 0,
            }
        )
        self.left_align_italic = self.workbook.add_format(
            {
                "font": "arial",
                "italic": True,
                "align": "left",
                "font_size": 8,
                "border": 0,
            }
        )
        self.left_align_italic_large = self.workbook.add_format(
            {
                "font": "arial",
                "italic": True,
                "align": "left",
                "font_size": 10,
                "border": 0,
            }
        )
        self.left_align_italic_large_ul = self.workbook.add_format(
            {
                "font": "arial",
                "italic": True,
                "align": "left",
                "underline": True,
                "font_size": 10,
                "border": 0,
            }
        )
        self.left_align_bold = self.workbook.add_format(
            {
                "font": "arial",
                "bold": True,
                "align": "left",
                "font_size": 8,
                "border": 0,
            }
        )
        self.left_align_bold_large = self.workbook.add_format(
            {
                "font": "arial",
                "bold": True,
                "align": "left",
                "font_size": 10,
                "border": 0,
            }
        )
        self.left_align_bold_larger = self.workbook.add_format(
            {
                "font": "arial",
                "bold": True,
                "align": "left",
                "font_size": 12,
                "border": 0,
            }
        )
        self.left_align = self.workbook.add_format(
            {"font": "arial", "align": "left", "font_size": 8, "border": 0}
        )
        self.left_align_large_wrap = self.workbook.add_format(
            {
                "font": "arial",
                "align": "left",
                "text_wrap": True,
                "font_size": 10,
                "border": 0,
            }
        )
        self.left_align_large_wrap_top = self.workbook.add_format(
            {
                "font": "arial",
                "align": "left",
                "valign": "top",
                "text_wrap": True,
                "font_size": 10,
                "border": 0,
            }
        )
        self.left_align_large = self.workbook.add_format(
            {"font": "arial", "align": "left", "font_size": 10, "border": 0}
        )
        self.right_align = self.workbook.add_format(
            {"font": "arial", "align": "right", "font_size": 8, "border": 0}
        )
        self.header_format = self.workbook.add_format(
            {
                "font": "arial",
                "bold": True,
                "italic": True,
                "text_wrap": True,
                "align": "left",
                "font_size": 8,
            }
        )
        self.header_format.set_bottom()
        self.header_format.set_top()

        self.title_format = self.workbook.add_format(
            {
                "font": "arial",
                "bold": True,
                "italic": False,
                "text_wrap": True,
                "align": "centre",
                "font_size": 12,
            }
        )
        self.section_heading = self.workbook.add_format(
            {
                "font": "arial",
                "bold": True,
                "italic": True,
                "text_wrap": True,
                "align": "left",
                "font_size": 11,
            }
        )

        self.footer_format = self.workbook.add_format(
            {
                "font": "arial",
                "align": "left",
                "font_size": 8,
            }
        )
        self.footer_format.set_top()
        self.merge_format = self.workbook.add_format(
            {"border": 1, "align": "center", "valign": "vcenter"}
        )

        self.date_format = self.workbook.add_format(
            {
                "num_format": "dd-mm-yyyy",
                "font": "arial",
                "align": "left",
                "font_size": 8,
                "border": 0,
            }
        )


def update_width(label: str, max_width):
    width = len(label)
    if width > max_width:
        max_width = width
    return max_width


def spacing(n_char=5):
    return " " * n_char


def indent(string, n_char=5):
    return spacing(n_char=n_char) + string


def write_planning_to_excel(excel_file, project, header_info, column_widths):
    with pd.ExcelWriter(excel_file, engine="xlsxwriter") as writer:
        try:
            projects_per_employee = project.tasks
        except AttributeError as err:
            raise AttributeError(
                f"{err}\nproject heeft helemaal geen tasks. Hier gaat what fout"
            )
        else:
            for projecten_employee in projects_per_employee:
                write_project_to_excel(
                    project=projecten_employee,
                    writer=writer,
                    sheet_name=projecten_employee.name,
                    header_info=header_info,
                    column_widths=column_widths,
                )


def write_project_to_excel(
    project,
    writer,
    sheet_name,
    header_info=None,
    column_widths=None,
    character_width=1.0,
):
    """
    Schrijf een multi index data frame naar Excel file met format

    Parameters
    ----------
    column_widths: dict
        Fix width of these columns.
    header_info: dict
        Information on the header
    project:
        Hoofdproject
    writer: obj
    sheet_name: str
        De sheet name
    character_width: float
        Width of 1 character. Default = 0.7

    """

    ExcelFormatter.header_style = None

    table_df = pd.DataFrame(data=[sheet_name])
    table_df.to_excel(
        writer, sheet_name=sheet_name, startrow=0, header=False, index=False
    )

    worksheet = writer.sheets[sheet_name]
    # worksheet.screen_gridlines = False
    workbook = writer.book

    wb = WorkBook(workbook=workbook)

    write_header(
        header_info=header_info,
        workbook=workbook,
        worksheet=worksheet,
        character_width=character_width,
        wb=wb,
        column_widths=column_widths,
    )

    row_index = 2
    level = 0

    _, level = write_project(
        project,
        header_info=header_info,
        workbook=workbook,
        worksheet=worksheet,
        character_width=character_width,
        wb=wb,
        row_index=row_index,
        level=level,
    )


def write_project(
    project, header_info, workbook, worksheet, character_width, wb, row_index, level
):
    _logger.debug("Writing project")
    col_index = 0
    resource_index = 0

    if level == 1:
        row_index += 1

    task_label = None
    for info_key, info_val in header_info.items():
        _logger.debug(f"Adding project {info_key}")
        columns_names = info_val["columns"]
        for column_key, column_name in columns_names.items():
            _logger.debug(f"Adding column {column_key}")
            try:
                label = getattr(project, column_key)
            except AttributeError:
                # als de kolom geen attribute heeft dan gewoon naar de volgende
                label = None
            if type(project) in (gantt.Task, gantt.Milestone):
                if column_key == "name":
                    task_label = label
                    label = None
                elif column_key == "task":
                    label = task_label
                elif column_key.startswith("employee"):
                    try:
                        employee = project.get_resources()[resource_index]
                    except IndexError:
                        label = None
                    else:
                        label = employee.name
                        resource_index += 1
                elif column_key == "period":
                    label = ""
                    try:
                        year_start = pd.Timestamp(project.start).year
                    except AttributeError:
                        year_start = ""
                    else:
                        year_start = str(year_start)[-2:]
                    label += f"{year_start}"
                    try:
                        quarter_start = pd.Timestamp(project.start).quarter
                    except AttributeError:
                        quarter_start = ""
                    else:
                        label += f"Q{quarter_start}"
                    try:
                        year_end = pd.Timestamp(project.stop).year
                    except AttributeError:
                        year_end = ""
                    else:
                        year_end = str(year_end)[-2:]
                    try:
                        quarter_end = pd.Timestamp(project.stop).quarter
                    except AttributeError:
                        pass
                    else:
                        if quarter_end != quarter_start or year_start != year_end:
                            label += f"/{year_end}Q{quarter_end}"
            if label is not None:
                _logger.debug(f"Writing {column_key} with {label}")
                try:
                    dummy = label.strftime("%d-%m-%Y")
                except AttributeError:
                    is_date = False
                else:
                    is_date = True
                    _logger.debug(f"Label is a date with format {dummy}")
                if col_index == 0 and level < 2:
                    formaat = wb.left_align_bold
                elif is_date:
                    formaat = wb.date_format
                else:
                    formaat = wb.left_align

                worksheet.write(row_index, col_index, label, formaat)
            col_index += 1

    row_index += 1

    try:
        tasks = project.tasks
    except AttributeError:
        _logger.debug("This is a task, so does not have tasks ")
    else:
        for task in tasks:
            level += 1
            row_index, level = write_project(
                task,
                header_info=header_info,
                workbook=workbook,
                worksheet=worksheet,
                character_width=character_width,
                wb=wb,
                row_index=row_index,
                level=level,
            )

    level -= 1
    return row_index, level


def write_header(header_info, workbook, worksheet, character_width, wb, column_widths):
    row_index = 0
    col_index = 0
    # begin met tabel nummer op eerste regel en title op regel 2
    for info_key, info_val in header_info.items():
        _logger.debug(f"Adding header for {info_key}")
        columns_names = info_val["columns"]
        title = info_val["title"]
        n_columns = len(columns_names.keys())
        if cell_color := info_val.get("color"):
            color = color_to_hex(cell_color)
        else:
            color = "black"

        merge_format = workbook.add_format(
            {
                "bold": True,
                "border": 6,
                "align": "center",
                "valign": "vcenter",
                "fg_color": color,
            }
        )
        if n_columns > 1:
            first_col = col_index
            last_col = col_index + n_columns - 1
            _logger.debug(
                f"Merging cells {first_col} - {last_col} at ro {row_index}: {title} {color}"
            )
            worksheet.merge_range(
                row_index, first_col, row_index, last_col, title, merge_format
            )
        else:
            _logger.debug(f"Writing cell {col_index}  at ro {row_index}: {title}")
            worksheet.write(row_index, col_index, title, merge_format)

        for column_key, column_name in columns_names.items():
            _logger.debug(f"Adding column {column_key}")
            worksheet.write(row_index + 1, col_index, column_name, wb.left_align_bold)

            column_width = len(column_name)
            if column_widths is not None:
                for col_key, col_width in column_widths.items():
                    if col_key == column_key:
                        column_width = col_width
            worksheet.set_column(col_index, col_index, column_width * character_width)
            col_index += 1



import pytest

from gantt_project_maker.main import check_if_items_are_available

__author__ = "Eelco van Vliet"
__copyright__ = "Eelco van Vliet"
__license__ = "MIT"


def test_check_if_items_are_available():
    """API Tests if check succeeds"""

    input_list = ["A", "B"]
    available_dict = {"A": 0, "B": 1, "C": 2}

    assert check_if_items_are_available(
        requested_items=input_list, available_items=available_dict, label="test1"
    )


def test_check_if_items_are_available_error():
    """API Tests if check files. Should raise a value error"""

    input_list = ["A", "B", "D"]
    available_dict = {"A": 0, "B": 1, "C": 2}
    with pytest.raises(ValueError):
        check_if_items_are_available(
            requested_items=input_list, available_items=available_dict, label="test2"
        )



import logging
from datetime import datetime
from pathlib import Path
from typing import Union
import copy

import dateutil.parser as dparse
from dateutil.parser import ParserError

import gantt_project_maker.gantt as gantt
from gantt_project_maker.colors import color_to_hex
from gantt_project_maker.excelwriter import write_planning_to_excel

SCALES = dict(
    daily=gantt.DRAW_WITH_DAILY_SCALE,
    weekly=gantt.DRAW_WITH_WEEKLY_SCALE,
    monthly=gantt.DRAW_WITH_MONTHLY_SCALE,
    quarterly=gantt.DRAW_WITH_QUARTERLY_SCALE,
)

_logger = logging.getLogger(__name__)


def get_nearest_saturday(date):
    """
    Get the nearest Saturday with respect to 'date'

    Parameters
    ----------
    date: datetime.date
        The reference date

    Returns
    -------
    datetime.date
        Nearest Saturday with respect to the reference date

    """
    d = date.toordinal()
    last = d - 6
    sunday = last - (last % 7)
    saturday = sunday + 6
    if d - saturday > 7 / 2:
        # de afstand tot vorige zaterdag is meer dan een halve week, dus de volgende zaterdag is dichter bij
        saturday += 7
    return date.fromordinal(saturday)


def parse_date(
    date_string: str, date_default: str = None, dayfirst=False
) -> datetime.date:
    """
    Lees de date_string en parse de datum

    Parameters
    ----------
    date_string: str
        Datum representatie
    date_default:
        Als de date_string None is deze default waarde genomen.
    dayfirst: bool
        Set the day first, e.g. 25-12-2023

    Returns
    -------
    datetime.date():
        Datum
    """
    if date_string is not None:
        date = dparse.parse(date_string.strip(), dayfirst=dayfirst).date()
    elif date_default is not None and isinstance(date_default, str):
        date = dparse.parse(date_default.strip(), dayfirst=dayfirst).date()
    else:
        date = date_default
    return date


def add_vacation_employee(employee: gantt.Resource, vacations: dict) -> dict:
    """
    Add the vacations of an employee

    Parameters
    ----------
    employee: gannt.Resource
        The employee for who you want to add the vacation
    vacations: dict
        A dictionary with items per vacation. Per vacation, you need a start and an end

    Returns
    -------
    dict:
        Dictionary with the vacations.
    """
    vacation_objects = dict()

    if vacations is not None:
        for vacation_key, vacation_properties in vacations.items():
            vacation_objects[vacation_key] = Vacation(
                vacation_properties["start"],
                vacation_properties.get("end"),
                employee=employee,
            )
    return vacation_objects


class StartEndBase:
    """
    Basis van alle classes met een begin- en einddatum.
    """

    def __init__(self, start: str, end: str = None, dayfirst=False):
        """
        Sla de datum stings op als datetime objecten

        Parameters
        ----------
        start: str
            Start date is mandatory
        end: str or None
            End date is optional.
        dayfirst: bool
            Use date with date first
        """
        self.start = parse_date(start, dayfirst=dayfirst)
        self.end = parse_date(end, dayfirst=dayfirst)


class Vacation(StartEndBase):
    def __init__(self, start, end=None, employee=None, dayfirst=False):
        super().__init__(start, end, dayfirst=dayfirst)

        if employee is None:
            self.pool = gantt
        else:
            self.pool = employee

        self.add_vacation()

    def add_vacation(self):
        """
        Add common or employee vacations
        """
        self.pool.add_vacations(self.start, self.end)


class Employee:
    def __init__(self, label, full_name=None, vacations=None, color=None):
        self.label = label
        self.full_name = full_name
        self.color = color
        self.resource = gantt.Resource(name=label, fullname=full_name, color=color)

        if vacations is not None:
            self.vacations = add_vacation_employee(
                employee=self.resource, vacations=vacations
            )
        else:
            self.vacations = None


class BasicElement(StartEndBase):
    def __init__(
        self,
        label,
        start=None,
        dependent_of=None,
        color=None,
        project_color=None,
        detail=False,
        display=True,
        dayfirst=False,
    ):
        super().__init__(start, start, dayfirst)
        if label is None:
            raise ValueError("Every task should have a label!")
        self.label = label
        self.detail = detail
        self.dependent_of = dependent_of
        self.color = color_to_hex(color)
        self.project_color = color_to_hex(project_color)
        self.display = display


class Task(BasicElement):
    def __init__(
        self,
        label,
        start=None,
        end=None,
        duration=None,
        employees=None,
        dependent_of=None,
        color=None,
        project_color=None,
        detail=False,
        display=True,
        dayfirst=True,
    ):
        super().__init__(
            label=label,
            start=start,
            dependent_of=dependent_of,
            color=color,
            project_color=project_color,
            detail=detail,
            display=display,
            dayfirst=dayfirst,
        )
        self.end = parse_date(end, dayfirst=dayfirst)
        self.duration = duration
        if self.end is None and self.duration is None:
            msg = (
                f"For a Task, next to a start date, either a end date or a duration needs to be specified. "
                f"None is given for task '{label}'"
            )
            raise ValueError(msg)
        if self.end is not None:
            if self.end < self.start:
                msg = f"End date {self.end} is before {self.start} for Task '{label}'. Please fix this"
                raise ValueError(msg)
        self.employees = employees

        self.element = self.add_task()

    def add_task(self) -> gantt.Task:
        if self.color is None:
            self.color = self.project_color
        task = gantt.Task(
            name=self.label,
            start=self.start,
            stop=self.end,
            duration=self.duration,
            depends_of=self.dependent_of,
            resources=self.employees,
            color=self.color,
        )
        return task


class Milestone(BasicElement):
    def __init__(
        self,
        label,
        start=None,
        dependent_of=None,
        color=None,
        project_color=None,
        detail=False,
        display=True,
        dayfirst=True,
    ):
        super().__init__(
            label=label,
            start=start,
            dependent_of=dependent_of,
            color=color,
            project_color=project_color,
            detail=detail,
            display=display,
            dayfirst=dayfirst,
        )

        self.element = self.add_milestone()

    def add_milestone(self) -> gantt.Milestone:
        element = gantt.Milestone(
            name=self.label,
            start=self.start,
            depends_of=self.dependent_of,
            color=self.color,
        )
        return element


class ProjectPlanner:
    def __init__(
        self,
        programma_title: str = None,
        vacations_title: str = None,
        programma_color: str = None,
        vacation_color: str = None,
        output_file_name: Path = None,
        planning_start: datetime = None,
        planning_end: datetime = None,
        weeks_margin_left: int = None,
        weeks_margin_right: int = None,
        today: datetime = None,
        dayfirst: bool = False,
        scale: str = None,
        period_info: dict = None,
        excel_info: dict = None,
        details: bool = None,
        filter_employees: list = None,
        save_svg_as_pdf: bool = False,
        collaps_tasks: bool = False,
    ):
        """

        Args:
            programma_title: str
                Main titel of the whole project
            vacations_title: str
                Titel of the vacations project
            programma_color: str
                First color of the bar
            output_file_name: str
                Base name of the output files
            planning_start: datetime
                Start of the program
            planning_end: datetime
                End of the program
            weeks_margin_left: int
                Shift the end of the planning so many weeks to the left without adding projects
            weeks_margin_right: int
                Shift the end of the planning so many weeks to the right without adding projects
            today: datetime
                Today's date
            dayfirst: bool
                Parse date with the day first
            scale: str
                Which scale is used for the output
            period_info: dict
               Information on the periods output
            excel_info: dict
                Information on the Excel output
            details: bool
                If true, include the details to the programs
            filter_employees: list
                If not None, only add task to which  employees in this list contribute
        """
        self.period_info = period_info
        self.planning_start = planning_start
        self.planning_end = planning_end
        self.date_today = today
        self.dayfirst = dayfirst
        self.scale = scale
        self.details = details
        self.save_svg_as_pdf = save_svg_as_pdf
        self.collaps_tasks = collaps_tasks
        if filter_employees:
            # if filter_employees are given, add them as a set
            self.filter_employees = set(filter_employees)
        else:
            self.filter_employees = None

        self.weeks_margin_left = weeks_margin_left
        self.weeks_margin_right = weeks_margin_right

        self.excel_info = excel_info

        if output_file_name is None:
            self.output_file_name = Path("gantt_projects.svg")
        else:
            self.output_file_name = Path(output_file_name)

        # Make the main project
        self.programma = gantt.Project(
            name=programma_title, color=color_to_hex(programma_color)
        )

        # Make the project to store all the vacations
        self.vacations_gantt = gantt.Project(
            name=vacations_title, color=color_to_hex(vacation_color)
        )

        self.project_tasks = dict()
        self.vacations = dict()
        self.employees = dict()
        self.tasks_and_milestones = dict()
        self.subprojects = dict()

    @staticmethod
    def add_global_information(
        fill="black", stroke="black", stroke_width=0, font_family="Verdana"
    ):
        gantt.define_font_attributes(
            fill=fill, stroke=stroke, stroke_width=stroke_width, font_family=font_family
        )

    def exporteer_naar_excel(self, excel_output_directory):
        """
        Write planning to an Excel file

        Parameters
        ----------
        excel_output_directory: Path
            Output directory of the Excel files
        """

        if self.excel_info is None:
            _logger.warning(
                "Cannot write excel! Please add Excel info to your settings file"
            )
        else:
            excel_output_directory.mkdir(exist_ok=True)
            excel_file = excel_output_directory / self.output_file_name.with_suffix(
                ".xlsx"
            )
            _logger.info(f"Exporting planning to {excel_file}")
            write_planning_to_excel(
                excel_file=excel_file,
                project=self.programma,
                header_info=self.excel_info["header"],
                column_widths=self.excel_info.get("column_widths"),
            )

    def get_dependency(self, key: str) -> gantt.Resource:
        """
        Search the object to which the dependency 'key' refers to

        Parameters
        ----------
        key: str
            Key of the dictionary to which the dependency refers to

        Returns
        -------
        gantt.Resource
            Object to which the  key refers to.
        """

        try:
            depends_of = self.tasks_and_milestones[key]
            if key in self.subprojects.keys():
                _logger.warning(
                    f"The dependency {key} occurs in both tasks en milestones"
                )
            _logger.debug(f"Dependent of task or milestone: {key}")
        except KeyError:
            try:
                depends_of = self.subprojects[key]
                _logger.debug(f"Dependent of project: {key}")
            except KeyError:
                msg = f"Dependency {key} does not exist"
                if self.filter_employees is not None:
                    # in case we are filtering on employees, some dependencies may be missing. Just give a warning
                    _logger.warning(msg)
                    depends_of = None
                else:
                    raise AssertionError(msg)

        return depends_of

    def get_employees(self, employees: Union[str, list]) -> list:
        """
        Turn a list of employees strings into a list of employees gannt.Resource objects

        Parameters
        ----------
        employees: list of str
            List of employees or, in case just one employee is given, a string


        Returns
        -------
        list:
            List of employees resource objects
        """

        employees_elements = list()
        if employees is not None:
            if isinstance(employees, str):
                _logger.debug(f"Adding employee: {employees}")
                employees_elements.append(self.employees[employees].resource)
            else:
                for employee in employees:
                    _logger.debug(f"Adding employee {employee}")
                    employees_elements.append(self.employees[employee].resource)
        return employees_elements

    def get_dependencies(self, dependencies: Union[str, dict]) -> list:
        """
        Retrieve all dependencies

        Parameters
        ----------
        dependencies: str or dict
            In case the dependency is a string, there is only one. This will be obtained from the dict.
            The dependencies may also be stored in a dict. In that case, we retrieve them per item

        Returns
        -------
        list:
            List of dependencies
        """

        dependency_elements = list()

        if dependencies is not None:
            if isinstance(dependencies, str):
                dependent_of = self.get_dependency(dependencies)
                dependency_elements.append(dependent_of)
            elif isinstance(dependencies, dict):
                for category, afhankelijk_items in dependencies.items():
                    for task_key in afhankelijk_items:
                        dependent_of = self.get_dependency(task_key)
                        dependency_elements.append(dependent_of)
            else:
                for afhankelijk_item in dependencies:
                    dependent_of = self.get_dependency(afhankelijk_item)
                    dependency_elements.append(dependent_of)

            return dependency_elements

    def add_vacations(self, vacations_info):
        """
        Add all the vacations
        """
        _logger.info("Adding general holidays")
        for v_key, v_prop in vacations_info.items():
            if v_prop.get("end") is not None:
                _logger.debug(
                    f"Vacation {v_key} from {v_prop['start']} to {v_prop.get('end')}"
                )
            else:
                _logger.debug(f"Vacation {v_key} at {v_prop['start']}")
            self.vacations[v_key] = Vacation(
                start=v_prop["start"], end=v_prop.get("end"), dayfirst=self.dayfirst
            )

    def add_employees(self, employees_info):
        """
        Add the employees with their vacations
        """
        _logger.info("Adding employees...")
        for w_key, w_prop in employees_info.items():
            _logger.debug(f"Adding {w_key} ({w_prop.get('name')})")
            full_name = w_prop.get("name")
            employee_vacations_info = w_prop.get("vacations")
            employee_color = w_prop.get("color")

            self.employees[w_key] = Employee(
                label=w_key,
                full_name=full_name,
                color=employee_color,
                vacations=employee_vacations_info,
            )

            # also add the vacations of this employee to the vacation gantt project in order to give an overview later
            employee_vacations = gantt.Project(
                name=full_name,
                color=employee_color,
                font=gantt.get_font_attributes(font_weight="bold", font_size="20"),
            )
            if employee_vacations_info is not None:
                for v_key, v_prop in employee_vacations_info.items():
                    vacation_name = v_prop.get("label", v_key)
                    vacation_start = v_prop.get("start")
                    vacation_end = v_prop.get("end")
                    if vacation_end is None:
                        vacation_duration = 1
                    else:
                        vacation_duration = None
                    vacation_task = Task(
                        label=vacation_name,
                        color=v_prop.get("color"),
                        start=vacation_start,
                        end=vacation_end,
                        duration=vacation_duration,
                        dayfirst=self.dayfirst,
                    )
                    employee_vacations.add_task(vacation_task.element)
            self.vacations_gantt.add_task(employee_vacations)

    def make_task_of_milestone(
        self,
        task_properties: dict = None,
        project_color=None,
    ) -> Union[Task, Milestone]:
        """
        Add all the general tasks and milestones

        Parameters
        ----------
        task_properties: dict
            Dictionary with tasks or milestones
        project_color: str
            Color of the parent project

        Returns
        -------
        Task or milestone

        """
        dependencies = self.get_dependencies(task_properties.get("dependent_of"))
        element_type = task_properties.get("type", "task")
        if element_type == "task":
            employees = self.get_employees(task_properties.get("employees"))
            _logger.debug(f"Voeg task {task_properties.get('label')} toe")
            task_or_milestone = Task(
                label=task_properties.get("label"),
                start=task_properties.get("start"),
                end=task_properties.get("end"),
                duration=task_properties.get("duration"),
                color=task_properties.get("color"),
                project_color=project_color,
                detail=task_properties.get("detail", False),
                employees=employees,
                dependent_of=dependencies,
                dayfirst=self.dayfirst,
            )
        elif element_type == "milestone":
            _logger.debug(f"Adding milestone {task_properties.get('label')} toe")
            task_or_milestone = Milestone(
                label=task_properties.get("label"),
                start=task_properties.get("start"),
                color=task_properties.get("color"),
                project_color=project_color,
                dependent_of=dependencies,
                dayfirst=self.dayfirst,
            )
        else:
            raise AssertionError("Type should be 'task' or 'milestone'")

        # add all the remain fields which are not required for the gantt charts but needed for the Excel output
        for task_key, task_value in task_properties.items():
            if not hasattr(task_or_milestone.element, task_key):
                if isinstance(task_value, str):
                    try:
                        _task_value = parse_date(
                            task_value, task_value, dayfirst=self.dayfirst
                        )
                    except ParserError:
                        _logger.debug(f"task {task_key} is not an date. No problem")
                    else:
                        _logger.debug(
                            f"Converted string {task_value} into datetime {_task_value}"
                        )
                        task_value = _task_value
                _logger.debug(f"Adding task {task_key} with value {task_value}")
                setattr(task_or_milestone.element, task_key, task_value)

        return task_or_milestone

    def add_tasks_and_milestones(
        self, tasks_and_milestones=None, tasks_and_milestones_info=None
    ):
        """
        Make all tasks en milestones
        """

        _logger.info("Voeg alle algemene tasks en mijlpalen toe")
        if tasks_and_milestones_info is not None:
            # The tasks are organised in modules, to peel of the first level
            tasks_en_mp = dict()
            for module_key, module_values in tasks_and_milestones_info.items():
                _logger.debug(f"Reading tasks of module {module_key}")
                for task_key, task_val in module_values.items():
                    _logger.debug(f"Processing task {task_key}")
                    if task_key in tasks_en_mp.keys():
                        msg = f"De task key {task_key} has been used before. Please pick another name"
                        _logger.warning(msg)
                        raise ValueError(msg)
                    if self.filter_employees is not None:
                        contributors = task_val.get("employees")
                        is_contributing = check_if_employee_in_contributing(
                            filter_employees=self.filter_employees,
                            contributing_employees=contributors,
                        )
                        if not is_contributing:
                            _logger.debug(
                                f"None of {contributors} are in {self.filter_employees}. Skipping"
                            )
                            continue
                    tasks_en_mp[task_key] = tasks_and_milestones_info[module_key][
                        task_key
                    ]
        else:
            tasks_en_mp = tasks_and_milestones

        for task_key, task_val in tasks_en_mp.items():
            _logger.debug(f"Processing task {task_key}")
            self.tasks_and_milestones[task_key] = self.make_task_of_milestone(
                task_properties=task_val
            )

    def make_projects(
        self,
        subprojects_info,
        subprojects_title,
        subprojects_selection,
        subprojects_color=None,
    ):
        """
        Make all projects
        """
        employee_color = color_to_hex(subprojects_color)

        projects_employee = gantt.Project(
            name=subprojects_title,
            color=employee_color,
            font=gantt.get_font_attributes(font_weight="bold", font_size="20"),
        )

        _logger.info(f"Add all projects of {subprojects_title}")
        for project_key, project_values in subprojects_info.items():
            project_name = project_values.get("title", project_key)
            project_name_global = project_values.get("title_collapsed", project_name)
            projects_employee_global = project_values.get("employees_collapsed")
            if projects_employee_global is not None and isinstance(
                projects_employee_global, str
            ):
                projects_employee_global = [projects_employee_global]

            _logger.info(f"Making project: {project_name}")

            project_color = color_to_hex(project_values.get("color"))

            _logger.debug("Creating project {}".format(project_name))
            project = gantt.Project(name=project_name, color=project_color)

            main_start_date = None
            main_end_date = None
            main_contributors = None

            # add all the other elements as attributes
            for p_key, p_value in project_values.items():
                if not hasattr(project, p_key):
                    setattr(project, p_key, p_value)

            if project_key in self.subprojects.keys():
                msg = f"project {project_key} already exists. Pick another name"
                _logger.warning(msg)
                raise ValueError(msg)

            self.subprojects[project_key] = project
            added_projects = list()

            if tasks := project_values.get("tasks"):
                if isinstance(tasks, list):
                    tasks_dict = {k: k for k in tasks}
                else:
                    tasks_dict = tasks

                for task_key, task_val in tasks_dict.items():
                    if isinstance(task_val, dict):
                        is_detail = task_val.get("detail", False)
                    else:
                        is_detail = False
                    if not self.details and is_detail:
                        # We hebben details op False staan en dit is een detail, dus sla deze task over.
                        _logger.info(
                            f"Skipping task {task_key} because it is set to be a detail"
                        )
                        continue

                    _logger.debug("Adding task {}".format(task_key))

                    is_detail = False

                    if isinstance(task_val, str):
                        try:
                            # de task een task of een milestone?
                            task_obj = self.tasks_and_milestones[task_val]
                            task = task_obj.element
                            is_detail = task_obj.detail
                        except KeyError:
                            try:
                                # de task een ander project?
                                task = self.subprojects[task_val]
                            except KeyError as err:
                                if not self.filter_employees:
                                    _logger.warning(f"{err}")
                                    raise
                                else:
                                    _logger.debug(f"{err}")
                                    continue
                    else:
                        task_obj = self.make_task_of_milestone(
                            task_properties=task_val, project_color=project_color
                        )
                        task = task_obj.element
                        is_detail = task_obj.detail

                    if task.color is None:
                        task.color = project_color

                    if self.filter_employees:
                        try:
                            contributors = task.employees
                        except AttributeError:
                            if isinstance(task, gantt.Task):
                                _logger.debug(
                                    "No employees found and this is a task. Skip it!"
                                )
                                continue
                            else:
                                _logger.debug(
                                    "No employees found, but can be a projects, so just add"
                                )
                        else:
                            is_contributing = check_if_employee_in_contributing(
                                filter_employees=self.filter_employees,
                                contributing_employees=contributors,
                            )
                            if not is_contributing:
                                _logger.debug(
                                    f"None of {contributors} are in {self.filter_employees}. Skipping"
                                )
                                continue

                    if not self.details and is_detail:
                        _logger.debug(f"skipping task {task_key} as it is a detail")
                    else:
                        # hier min en max data bijhouden
                        if not self.collaps_tasks or isinstance(task, gantt.Project):
                            added_projects.append(project_key)
                            project.add_task(task)

                        if projects_employee_global is None:
                            main_contributors = get_contributors_task(
                                task,
                                main_contributors,
                            )
                        else:
                            main_contributors = get_contributors_from_resources(
                                projects_employee_global,
                                main_contributors,
                                self.employees,
                            )

                        # each project with task is stored as a main project with the project begin and end
                        if main_start_date is None:
                            main_start_date = task.start_date()
                            main_end_date = task.end_date()

                        if task.end_date() > main_end_date:
                            main_end_date = task.end_date()

            # every project with tasks will be a course project plan as well
            if self.collaps_tasks and project_key not in added_projects:
                if main_contributors is not None:
                    main_contributors = list(set(main_contributors))
                main_task = gantt.Task(
                    name=project_name_global,
                    start=main_start_date,
                    stop=main_end_date,
                    resources=main_contributors,
                )
                if project is not None:
                    project.add_task(main_task)

            self.subprojects[project_key] = project
            if project_key in subprojects_selection:
                # hier project zonder taken toevoegen
                projects_employee.add_task(project)

        # add now all projects of the employee to the program
        self.programma.add_task(projects_employee)

    def write_planning(
        self,
        planning_output_directory,
        resource_output_directory,
        vacations_output_directory,
        write_resources=False,
        write_vacations=False,
        periods=None,
    ):
        """
        Write the planning to the output definitions

        Parameters
        ----------
        write_resources: bool
            Write the resources file as well (next to the gantt charts which is always written)
        write_vacations: bool
            Write the vacations file as well
        planning_output_directory: Path
            Output directory of the svg files of the planning
        resource_output_directory: Path
            Output directory of the svg files of the resources
        vacations_output_directory: Path
            Output directory of the svg files of the vacations
        periods: list
           Periods we want to add. If None, add all periods
        """

        directories = {
            "tasks": planning_output_directory,
            "resources": resource_output_directory,
            "vacations": vacations_output_directory,
        }
        svg42pdf = None
        # make output directories. Vacations can be made per period, so do later
        directories["tasks"].mkdir(parents=True, exist_ok=True)
        if write_resources:
            directories["resources"].mkdir(parents=True, exist_ok=True)

        for period_key, period_prop in self.period_info.items():
            if periods is not None and period_key not in periods:
                _logger.debug(f"Employee {period_key} is skipped")
                continue

            file_names = dict()
            leading_suffix = [period_key]
            if self.collaps_tasks:
                leading_suffix += ["collapsed"]

            write_vacations_this_period = period_prop.get("export_vacations", False)

            for file_suffix in directories.keys():
                file_names[file_suffix] = extend_suffix(
                    self.output_file_name, extensions=leading_suffix + [file_suffix]
                )
                file_names[file_suffix] = (
                    directories[file_suffix] / file_names[file_suffix]
                )

            weeks_margin_left = period_prop.get(
                "weeks_margin_left", self.weeks_margin_left
            )
            weeks_margin_right = period_prop.get(
                "weeks_margin_right", self.weeks_margin_right
            )

            scale = period_prop.get("scale")
            if scale is not None:
                scale = SCALES[scale]
            else:
                scale = self.scale

            start = parse_date(
                period_prop.get("planning_start"),
                self.planning_start,
                dayfirst=self.dayfirst,
            )
            end = parse_date(
                period_prop.get("planning_end"),
                self.planning_end,
                dayfirst=self.dayfirst,
            )

            today = parse_date(
                period_prop.get("today"), self.date_today, dayfirst=self.dayfirst
            )
            if today is not None and scale != SCALES["daily"]:
                # For any scale other than  daily, the today-line is drawn only at Saturdays
                _logger.debug("Change the date to the nearest Saturday")
                _today = today
                today = get_nearest_saturday(today)
                if today != _today:
                    _logger.debug(
                        f"Changing the date today {_today} into the the nearest Saturday {today}"
                    )

            # the planning is a collection of all the projects
            file_name = file_names["tasks"]
            _logger.info(
                f"Writing project starting at {start} and ending at {end} with a scale {scale} to {file_name}"
            )
            self.programma.make_svg_for_tasks(
                filename=file_name,
                start=start,
                end=end,
                margin_left=weeks_margin_left,
                margin_right=weeks_margin_right,
                scale=scale,
                today=today,
            )

            if self.save_svg_as_pdf:
                try:
                    import svg42pdf
                except ImportError as err:
                    _logger.warning(f"{err}\nFailed writing pdf because svg42pdf is")
                else:
                    pdf_file_name = file_name.with_suffix(".pdf")
                    _logger.info(f"Saving as {pdf_file_name}")
                    svg42pdf.svg42pdf(
                        svg_fn=file_name.as_posix(),
                        pdf_fn=pdf_file_name.as_posix(),
                        method="any",
                    )

            if write_resources:
                file_name_resources = file_names["resources"]
                _logger.info(
                    f"Write resources of {start} to {end} with scale {scale} to {file_name_resources}"
                )
                try:
                    self.programma.make_svg_for_resources(
                        filename=file_name_resources.as_posix(),
                        start=start,
                        end=end,
                        scale=SCALES["daily"],
                        today=today,
                    )
                except TypeError as err:
                    _logger.warning(err)
                    _logger.warning(
                        "Failed writing the resources with the above error message. Please check your "
                        "employee's input data"
                    )
                else:
                    if self.save_svg_as_pdf and svg42pdf is not None:
                        pdf_file_name_res = file_name_resources
                        svg42pdf.svg42pdf(
                            svg_fn=file_name_resources.as_posix(),
                            pdf_fn=pdf_file_name_res.as_posix(),
                            method="any",
                        )

            if write_vacations or write_vacations_this_period:
                file_name_vacations = file_names["vacations"]
                vacations_output_directory.mkdir(exist_ok=True, parents=True)
                _logger.info(f"Writing vacation file {file_name_vacations}")
                self.vacations_gantt.make_svg_for_tasks(
                    filename=file_name_vacations,
                    start=start,
                    end=end,
                    margin_left=weeks_margin_left,
                    margin_right=weeks_margin_right,
                    scale=scale,
                    today=today,
                )
                if self.save_svg_as_pdf and svg42pdf is not None:
                    pdf_file_name_vac = file_name_vacations.with_suffix(".pdf")
                    svg42pdf.svg42pdf(
                        svg_fn=file_name_vacations.as_posix(),
                        pdf_fn=pdf_file_name_vac.as_posix(),
                        method="any",
                    )

            _logger.debug("Done")


def check_if_employee_in_contributing(
    filter_employees: list, contributing_employees: list
) -> bool:
    """
    Check if any of the employees given in filter_employees is in contributing to this taks
    """

    is_contributing = False

    if contributing_employees:
        if isinstance(contributing_employees, str):
            contributing_employees = set(list([contributing_employees]))
        else:
            contributing_employees = set(contributing_employees)
        if contributing_employees.intersection(filter_employees):
            _logger.debug(
                f"Contributing employees {contributing_employees} not in filter list"
            )
            is_contributing = True

    return is_contributing


def extend_suffix(output_filename: Path, extensions: Union[list, str]):
    """
    Add an extra suffix to the base filename

    Parameters
    ----------
    output_filename: Path
        Base filename
    extensions: str or list
        Extra suffixes to add

    Returns
    -------
    Path:
        New filename with extra suffix

    """
    suffix = output_filename.suffix
    if isinstance(extensions, str):
        extensions = [extensions]
    output_filename = Path(
        "_".join([output_filename.with_suffix("").as_posix()] + extensions)
    ).with_suffix(suffix)

    return output_filename


def get_contributors_task(task, contributors):
    """
    get of the contributors of this Task based on the task resources

    Parameters
    ----------
    task
    contributors

    Returns
    -------
    list:
        contributors

    """
    try:
        task_resources = task.resources
    except AttributeError:
        pass
    else:
        if task_resources:
            for employee in task_resources:
                if contributors is None:
                    contributors = [employee]
                else:
                    contributors.append(employee)
    return contributors


def get_contributors_from_resources(
    projects_employee_global, contributors, all_resources
):
    """
    get the contributors of this Task based on the global resources

    Parameters
    ----------
    projects_employee_global
    contributors
    all_resources: list

    Returns
    -------
    list:
        contributors

    """
    for emp_key, emp_val in all_resources.items():
        if emp_key in projects_employee_global:
            if contributors is None:
                contributors = [emp_val.resource]
            else:
                contributors.append(emp_val.resource)
    return contributors



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
gantt.py - version and date, see below

This is a python class to create gantt chart using SVG


Author : Alexandre Norman - norman at xael.org

Contributors:

* Sébastien NOBILI - pipoprods at free.fr


Licence : GPL v3 or any later version


This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""


import codecs
import datetime
import io
import logging
import re
import sys

import dateutil.relativedelta
import svgwrite

# original author: Alexandre Norman (norman at xael.org)
# modified by Eelco van Vliet

# conversion from mm/cm to pixel is done by ourselves as firefox seems
# to have a bug for big numbers...
# 3.543307 is for conversion from mm to pt units !
mm = 3.543307
cm = 35.43307

COLOR_OVERCHARGE_DEFAULT = "#AA0000"
COLOR_VACATION_DEFAULT = "#008000"
COLOR_RESOURCE_DEFAULT = "#c5f0eb"


class _my_svgwrite_drawing_wrapper(svgwrite.Drawing):
    """
    Hack for beeing able to use a file descriptor as filename
    """

    def save(self, width="100%", height="100%"):
        """Write the XML string to **filename**."""
        test = False
        import io

        # Fix height and width
        self["height"] = height
        self["width"] = width

        this_file_type = type(self.filename)

        test = this_file_type == io.TextIOWrapper

        if test:
            self.write(self.filename)
        else:
            with io.open(str(self.filename), mode="w", encoding="utf-8") as stream:
                self.write(stream)


############################################################################

# noinspection PyTypeChecker
__LOG__: logging.Logger = None

############################################################################

DRAW_WITH_DAILY_SCALE = "d"
DRAW_WITH_WEEKLY_SCALE = "w"
DRAW_WITH_MONTHLY_SCALE = "m"
DRAW_WITH_QUARTERLY_SCALE = "q"

############################################################################

# Unworked days (0: Monday ... 6: Sunday)
NOT_WORKED_DAYS = [5, 6]


def define_not_worked_days(list_of_days):
    """
    Define specific days off

    Keyword arguments:
    list_of_days -- list of integer (0: Monday ... 6: Sunday) - default [5, 6]
    """
    global NOT_WORKED_DAYS
    NOT_WORKED_DAYS = list_of_days
    return


def _not_worked_days():
    """
    Returns list of days off (0: Monday ... 6: Sunday)
    """
    global NOT_WORKED_DAYS
    return NOT_WORKED_DAYS


############################################################################

FONT_ATTR = {
    "fill": "black",
    "stroke": "black",
    "stroke_width": 0,
    "font_family": "Verdana",
    "font_size": 15,
    "font_weight": "normal",
}


def define_font_attributes(
    fill="black",
    stroke="black",
    stroke_width=0,
    font_family="Verdana",
    font_weight="normal",
    font_size=15,
):
    """
    Define font attributes

    Keyword arguments:
    fill -- fill - default 'black'
    stroke -- stroke - default 'black'
    stroke_width -- stroke width - default 0
    font_family -- font family - default 'Verdana'
    """
    global FONT_ATTR

    FONT_ATTR = {
        "fill": fill,
        "stroke": stroke,
        "stroke_width": stroke_width,
        "font_family": font_family,
        "font_weight": font_weight,
        "font_size": font_size,
    }

    return


def _font_attributes():
    """
    Return dictionary of font attributes
    Example :
    FONT_ATTR = {
      'fill': 'black',
      'stroke': 'black',
      'stroke_width': 0,
      'font_family': 'Verdana',
    }
    """
    global FONT_ATTR
    return FONT_ATTR


def get_font_attributes(
    fill=None,
    stroke=None,
    stroke_width=None,
    font_family=None,
    font_weight=None,
    font_size=None,
):
    """
    Return dictionary of font attributes
    """
    global FONT_ATTR

    font_attributes = FONT_ATTR.copy()
    if fill is not None:
        font_attributes["fill"] = fill
    if stroke is not None:
        font_attributes["stroke"] = stroke
    if stroke_width is not None:
        font_attributes["stroke_width"] = stroke_width
    if font_family is not None:
        font_attributes["font_family"] = font_family
    if font_weight is not None:
        font_attributes["font_weight"] = font_weight
    if font_size is not None:
        font_attributes["font_size"] = font_size

    return font_attributes


############################################################################


# list of vacations as datetime (non-worked days)
VACATIONS = []


############################################################################


def add_vacations(start_date, end_date=None):
    """
    Add vacations to a resource beginning at [start_date] to [end_date]
    (included). If [end_date] is not defined, vacation will be for [start_date]
    day only

    Keyword arguments:
    start_date -- datetime.date beginning of vacation
    end_date -- datetime.date end of vacation
    """
    __LOG__.debug(
        "** add_vacations {0}".format({"start_date": start_date, "end_date": end_date})
    )

    global VACATIONS

    if end_date is None:
        if start_date not in VACATIONS:
            VACATIONS.append(start_date)
    else:
        while start_date <= end_date:
            if start_date not in VACATIONS:
                VACATIONS.append(start_date)

            start_date += datetime.timedelta(days=1)

    __LOG__.debug(
        "** add_vacations {0}".format(
            {"start_date": start_date, "end_date": end_date, "vac": VACATIONS}
        )
    )

    return


############################################################################


def init_log_to_sysout(level=logging.INFO):
    """
    Init global variable __LOG__ used for logging purpose

    Keyword arguments:
    level -- logging level (from logging.debug to logging.critical)
    """
    global __LOG__
    logger = logging.getLogger("Gantt")
    logger.setLevel(level)
    fh = logging.StreamHandler()
    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
    fh.setFormatter(formatter)
    logger.addHandler(fh)
    __LOG__ = logging.getLogger("Gantt")
    return


def _flatten(nested_list, list_types=(list, tuple)):
    """
    Return a flatten list from a list like [1,2,[4,5,1]]
    """
    list_type = type(nested_list)
    nested_list = list(nested_list)
    i = 0
    while i < len(nested_list):
        while isinstance(nested_list[i], list_types):
            if not nested_list[i]:
                nested_list.pop(i)
                i -= 1
                break
            else:
                nested_list[i : i + 1] = nested_list[i]
        i += 1
    return list_type(nested_list)


############################################################################
class GroupOfResources:
    """
    Class for grouping resources
    """

    def __init__(self, name, fullname=None):
        """
        Init a group of resources

        Keyword arguments:
        name -- name given to the resource (id)
        fullname -- long name given to the resource
        """
        __LOG__.debug("** GroupOfResources::__init__ {0}".format({"name": name}))
        self.name = name
        self.vacations = []
        if fullname is not None:
            self.fullname = fullname
        else:
            self.fullname = name

        self.resources = []

        self.tasks = []
        return

    def add_resource(self, resource):
        """
        Add a resource to the group of resources

        Keyword arguments:
        resource -- Resource object
        """
        if resource not in self.resources:
            self.resources.append(resource)
            resource.add_group(self)
        return

    def add_vacations(self, dfrom, dto=None):
        """
        Add vacations to a resource beginning at [dfrom] to [dto] (included). If
        [dto] is not defined, vacation will be for [dfrom] day only

        Keyword arguments:
        dfrom -- datetime.date beginning of vacation
        dto -- datetime.date end of vacation
        """
        __LOG__.debug(
            "** Resource::add_vacations {0}".format(
                {"name": self.name, "dfrom": dfrom, "dto": dto}
            )
        )
        if dto is None:
            self.vacations.append((dfrom, dfrom))
        else:
            self.vacations.append((dfrom, dto))
        return

    def nb_elements(self):
        """
        Returns the number of resources
        """
        __LOG__.debug(
            "** GroupOfResources::nb_elements ({0})".format({"name": self.name})
        )
        return len(self.resources)

    def is_available(self, date):
        """
        Returns True if any resource is available at given date, False if not.
        Availibility is tasks from the global VACATIONS and resource's ones.

        Keyword arguments:
        date -- datetime.date day to look for
        """
        # Global VACATIONS
        if date in VACATIONS:
            __LOG__.debug(
                "** GroupOfResources::is_available {0} : False (global vacation)".format(
                    {"name": self.name, "date": date}
                )
            )
            return False

        # Group vacations
        for h in self.vacations:
            dfrom, dto = h
            if dfrom <= date <= dto:
                __LOG__.debug(
                    "** GroupOfResources::is_available {0} : False (group vacation)".format(
                        {"name": self.name, "date": date}
                    )
                )
                return False

        # Test if at least one resource is available
        for r in self.resources:
            if r.is_available(date):
                __LOG__.debug(
                    "** GroupOfResources::is_available {0} : True {1}".format(
                        {"name": self.name, "date": date}, r.name
                    )
                )
                return True

        __LOG__.debug(
            "** GroupOfResources::is_available {0} : False".format(
                {"name": self.name, "date": date}
            )
        )
        return False

    def add_task(self, task):
        """
        Tell the resource that we have assigned a task

        Keyword arguments:
        task -- Task object
        """
        if task not in self.tasks:
            self.tasks.append(task)
        return

    def search_for_task_conflicts(self, all_tasks=False):
        """
        Returns a dictionary of all days (datetime.date) containing for each
        overcharged day the list of task for this day.

        It examines all resources member and group tasks.

        Keyword arguments:
        all_tasks -- if True return all tasks for all days, not just overcharged days
        """
        # Get for each resource
        affected_days = {}
        for r in self.resources:
            ad = r.search_for_task_conflicts(all_tasks=True)
            for d in ad:
                try:
                    affected_days[d].append(ad[d])
                except KeyError:
                    affected_days[d] = [ad[d]]

        # inspect project
        for t in self.tasks:
            cday = t.start_date()
            while cday <= t.end_date():
                if cday.weekday() not in _not_worked_days():
                    try:
                        affected_days[cday].append(t.fullname)
                    except KeyError:
                        affected_days[cday] = [t.fullname]

                cday += datetime.timedelta(days=1)

        # compile everything
        overcharged_days = {}
        ke = list(affected_days.keys())
        ke.sort()
        for d in ke:
            affected_days[d] = _flatten(affected_days[d])
            if all_tasks:
                overcharged_days[d] = affected_days[d]

            elif len(affected_days[d]) > self.nb_elements():
                overcharged_days[d] = affected_days[d]
                __LOG__.warning(
                    '** GroupOfResources "{2}" has more than {3} tasks on day {0} / {1}'.format(
                        d, affected_days[d], self.name, self.nb_elements()
                    )
                )

        return overcharged_days

    def is_vacant(self, from_date, to_date):
        """
        Check if any resource from the group is unallocated between for a given timeframe.
        Returns a list of available ressource name.

        Keyword arguments:
        from_date -- first day
        to_date --  last day
        """
        availables = []
        for r in self.resources:
            if len(r.is_vacant(from_date, to_date)) > 0:
                availables.append(r.name)

        return availables


############################################################################


class Resource:
    """
    Class for handling resources assigned to tasks
    """

    def __init__(self, name, fullname=None, color=None):
        """
        Init a resource

        Keyword arguments:
        name -- name given to the resource (id)
        fullname -- long name given to the resource
        color -- color used to represent the resource in the resources overview
        """
        __LOG__.debug("** Resource::__init__ {0}".format({"name": name}))
        self.name = name
        self.color = color
        if fullname is not None:
            self.fullname = fullname
        else:
            self.fullname = name

        self.vacations = []
        self.member_of_groups = []

        self.tasks = []
        return

    def add_vacations(self, dfrom, dto=None):
        """
        Add vacations to a resource beginning at [dfrom] to [dto] (included). If
        [dto] is not defined, vacation will be for [dfrom] day only

        Keyword arguments:
        dfrom -- datetime.date beginning of vacation
        dto -- datetime.date end of vacation
        """
        __LOG__.debug(
            "** Resource::add_vacations {0}".format(
                {"name": self.name, "dfrom": dfrom, "dto": dto}
            )
        )
        if dto is None:
            self.vacations.append((dfrom, dfrom))
        else:
            self.vacations.append((dfrom, dto))
        return

    def nb_elements(self):
        """
        Returns the number of resources, 1 here
        """
        __LOG__.debug("** Resource::nb_elements ({0})".format({"name": self.name}))
        return 1

    def is_available(self, date):
        """
        Returns True if the resource is available at given date, False if not.
        Availibility is tasks from the global VACATIONS and resource's ones.

        Keyword arguments:
        date -- datetime.date day to look for
        """
        # global VACATIONS
        if date in VACATIONS:
            __LOG__.debug(
                "** Resource::is_available {0} : False (global vacation)".format(
                    {"name": self.name, "date": date}
                )
            )
            return False

        # GroupOfResources vacation
        for g in self.member_of_groups:
            for h in g.vacations:
                dfrom, dto = h
                if date >= dfrom and date <= dto:
                    __LOG__.debug(
                        "** Resource::is_available {0} : False (Group {1})".format(
                            {"name": self.name, "date": date}, g.name
                        )
                    )
                    return False

        # Resource vacation
        for h in self.vacations:
            dfrom, dto = h
            if date >= dfrom and date <= dto:
                __LOG__.debug(
                    "** Resource::is_available {0} : False".format(
                        {"name": self.name, "date": date}
                    )
                )
                return False
        __LOG__.debug(
            "** Resource::is_available {0} : True".format(
                {"name": self.name, "date": date}
            )
        )
        return True

    def add_group(self, groupofresources):
        """
        Tell the resource it belongs to a GroupOfResources

        Keyword arguments:
        groupofresources -- GroupOfResources
        """
        if groupofresources not in self.member_of_groups:
            self.member_of_groups.append(groupofresources)
        return

    def add_task(self, task):
        """
        Tell the resource that we have assigned a task

        Keyword arguments:
        task -- Task object
        """
        if task not in self.tasks:
            self.tasks.append(task)
        return

    def search_for_task_conflicts(self, all_tasks=False):
        """
        Returns a dictionary of all days (datetime.date) containing for each
        overcharged day the list of task for this day.

        Keyword arguments:
        all_tasks -- if True return all tasks for all days, not just overcharged days
        """
        affected_days = {}
        for t in self.tasks:
            try:
                task_start_date = t.start_date()
            except TypeError as err:
                __LOG__.warning(err)
                __LOG__.warning(
                    f"Could not get initial start date for task {t.fullname}. Is is properly defined?"
                )
                raise
            try:
                task_end_date = t.end_date()
            except TypeError as err:
                __LOG__.warning(err)
                __LOG__.warning(
                    f"Could not get end date for task {t.fullname}. Is is properly defined?"
                )
                raise
            try:
                while task_start_date <= task_end_date:
                    if task_start_date.weekday() not in _not_worked_days():
                        try:
                            affected_days[task_start_date].append(t.fullname)
                        except KeyError:
                            affected_days[task_start_date] = [t.fullname]

                    task_start_date += datetime.timedelta(days=1)
            except TypeError as err:
                __LOG__.warning(err)
                __LOG__.warning(
                    f"Failing for task {t.fullname} with {task_start_date} (init {t.start_date()} and {t.end_date()}"
                )
                raise

        # return all
        if all_tasks:
            return affected_days

        # compile only overcharge
        overcharged_days = {}
        ke = list(affected_days.keys())
        ke.sort()
        for d in ke:
            if len(affected_days[d]) > 1:
                overcharged_days[d] = affected_days[d]
                __LOG__.warning(
                    '** Resource "{2}" has more than one task on day {0} / {1}'.format(
                        d, affected_days[d], self.name
                    )
                )

        return overcharged_days

    def is_vacant(self, from_date, to_date):
        """
        Check if the resource is unallocated between for a given timeframe.
        Returns True if the resource is free, False otherwise

        Keyword arguments:
        from_date -- first day
        to_date --  last day
        """
        non_vacant_days = self.search_for_task_conflicts(all_tasks=True)
        cday = from_date
        while cday <= to_date:
            if cday.weekday() not in _not_worked_days():
                if not self.is_available(cday):
                    __LOG__.debug(
                        '** Ressource "{0}" is not available on day {1} (vacation)'.format(
                            self.name, cday
                        )
                    )
                    return []
                if cday in non_vacant_days:
                    __LOG__.debug(
                        '** Ressource "{0}" is not available on day {1} (other task : {2})'.format(
                            self.name, cday, non_vacant_days[cday]
                        )
                    )
                    return []

            cday += datetime.timedelta(days=1)
        return [self.name]


############################################################################


class Task:
    """
    Class for manipulating Tasks
    """

    def __init__(
        self,
        name,
        start=None,
        stop=None,
        duration: int = None,
        depends_of=None,
        resources=None,
        percent_done=0,
        color=None,
        fullname=None,
        display=True,
        state="",
    ):
        """
        Initialize task object. Two of start, stop or duration may be given.
        This task can rely on other task and will be completed with resources.
        If percent done is given, a progress bar will be included on the task.
        If color is specified, it will be used for the task.

        Keyword arguments:
        name -- name of the task (id)
        fullname -- long name given to the resource
        start -- datetime.date, first day of the task, default None
        stop -- datetime.date, last day of the task, default None
        duration -- int, duration of the task, default None
        depends_of -- list of Task which are parents of this one, default None
        resources -- list of Resources assigned to the task, default None
        percent_done -- int, percent of achievement, default 0
        color -- string, html color, default None
        display -- boolean, display this task, default True
        state -- string, state of the task
        """
        __LOG__.debug(
            "** Task::__init__ {0}".format(
                {
                    "name": name,
                    "start": start,
                    "stop": stop,
                    "duration": duration,
                    "depends_of": depends_of,
                    "resources": resources,
                    "percent_done": percent_done,
                }
            )
        )
        self.name = name
        if fullname is not None:
            self.fullname = fullname
        else:
            self.fullname = name

        self.start = start
        self.stop = stop
        self.duration: int = duration
        self.color = color
        self.display = display
        self.state = state

        ends = (self.start, self.stop, self.duration)
        nonecount = 0
        for e in ends:
            if e is None:
                nonecount += 1

        # check limits (2 must be set on 4) or scheduling is defined by duration and dependencies
        if nonecount != 1 and (self.duration is None or depends_of is None):
            __LOG__.error(
                '** Task "{1}" must be defined by two of three limits ({0})'.format(
                    {"start": self.start, "stop": self.stop, "duration": self.duration},
                    fullname,
                )
            )

        if type(depends_of) is type([]):
            self.depends_of = depends_of
        elif depends_of is not None:
            self.depends_of = [depends_of]
        else:
            self.depends_of = None

        self.resources = resources
        self.percent_done = percent_done
        self.drawn_x_begin_coord = None
        self.drawn_x_end_coord = None
        self.drawn_y_coord = None
        self._cache_start_date = None
        self._cache_end_date = None

        # tell each resource we have
        # assigned a new task
        if resources is not None:
            for r in resources:
                r.add_task(self)

        return

    def add_depends(self, depends_of):
        """
        Adds dependency to a task

        Keyword arguments:
        depends_of -- list of Task which are parents of this one
        """
        if type(depends_of) is type([]):
            if self.depends_of is None:
                self.depends_of = depends_of
            else:
                for d in depends_of:
                    self.depends_of.append(d)
        else:
            if self.depends_of is None:
                self.depends_of = depends_of
            else:
                self.depends_of.append(depends_of)

        return

    def start_date(self):
        """
        Returns the first day of the task, either the one which was given at
        task creation or the one calculated after checking dependencies
        """
        if self._cache_start_date is not None:
            return self._cache_start_date

        __LOG__.debug("** Task::start_date ({0})".format(self.name))
        if self.start is not None:
            # start date set, calculate beginning
            if self.depends_of is None:
                # depends on nothing... start date is start
                # __LOG__.debug('*** Do not depend of other task')
                start = self.start
                while start.weekday() in _not_worked_days() or start in VACATIONS:
                    start = start + datetime.timedelta(days=1)

                if start > self.start:
                    __LOG__.warning(
                        '** Due to vacations, Task "{0}", will not start on date {1} but {2}'.format(
                            self.fullname, self.start, start
                        )
                    )

                self._cache_start_date = start
                return self._cache_start_date
            else:
                # depends on other task, start date could vary
                # __LOG__.debug('*** Do depend of other tasks')
                start = self.start
                while start.weekday() in _not_worked_days() or start in VACATIONS:
                    start = start + datetime.timedelta(days=1)

                prev_task_end = start
                for t in self.depends_of:
                    if isinstance(t, Milestone):
                        if t.end_date() >= prev_task_end:
                            prev_task_end = t.end_date()
                    elif isinstance(t, Task):
                        if t.end_date() >= prev_task_end:
                            prev_task_end = t.end_date() + datetime.timedelta(days=1)

                while (
                    prev_task_end.weekday() in _not_worked_days()
                    or prev_task_end in VACATIONS
                ):
                    prev_task_end = prev_task_end + datetime.timedelta(days=1)

                if prev_task_end > self.start:
                    __LOG__.warning(
                        '** Due to dependencies, Task "{0}", will not start on date {1} but {2}'.format(
                            self.fullname, self.start, prev_task_end
                        )
                    )

                self._cache_start_date = prev_task_end
                return self._cache_start_date

        elif self.duration is None:  # start and stop fixed
            current_day = self.start
            # check depends
            if self.depends_of is not None:
                prev_task_end = self.depends_of[0].end_date()
                for t in self.depends_of:
                    if isinstance(t, Milestone):
                        if t.end_date() > prev_task_end:
                            prev_task_end = t.end_date() - datetime.timedelta(days=1)
                    elif isinstance(t, Task):
                        if t.end_date() > prev_task_end:
                            prev_task_end = t.end_date()
                    # if t.end_date() > prev_task_end:
                    #     #__LOG__.debug('*** latest one {0} which end on {1}'.format(t.name, t.end_date()))
                    #     prev_task_end = t.end_date()
                if prev_task_end > current_day:
                    depend_start_date = prev_task_end
                else:
                    start = self.start
                    while start.weekday() in _not_worked_days() or start in VACATIONS:
                        start = start + datetime.timedelta(days=1)
                    depend_start_date = start

                    if depend_start_date > current_day:
                        __LOG__.error(
                            '** Due to dependencies, Task "{0}", could not be finished on time (should start as last '
                            "on {1} but will start on {2})".format(
                                self.fullname, current_day, depend_start_date
                            )
                        )
                    self._cache_start_date = depend_start_date
            else:
                # should be first day of start...
                self._cache_start_date = current_day

            return self._cache_start_date

        elif (
            self.duration is not None
            and self.depends_of is not None
            and self.stop is None
        ):  # duration and dependencies fixed
            prev_task_end = self.depends_of[0].end_date()
            for t in self.depends_of:
                if isinstance(t, Milestone):
                    if t.end_date() > prev_task_end:
                        prev_task_end = t.end_date() - datetime.timedelta(days=1)
                elif isinstance(t, Task):
                    if t.end_date() > prev_task_end:
                        prev_task_end = t.end_date()
                # if t.end_date() > prev_task_end:
                #     __LOG__.debug('*** latest one {0} which end on {1}'.format(t.name, t.end_date()))
                #     prev_task_end = t.end_date()

            start = prev_task_end + datetime.timedelta(days=1)

            while start.weekday() in _not_worked_days() or start in VACATIONS:
                start = start + datetime.timedelta(days=1)

            # should be first day of start...
            self._cache_start_date = start

        elif self.start is None and self.stop is not None:  # stop and duration fixed
            # start date not setted, calculate from end_date + depends
            current_day = self.stop
            real_duration = 0
            duration = self.duration
            while duration > 0:
                if not (
                    current_day.weekday() in _not_worked_days()
                    or current_day in VACATIONS
                ):
                    real_duration = real_duration + 1
                    duration -= 1
                else:
                    real_duration = real_duration + 1

                current_day = self.stop - datetime.timedelta(days=real_duration)
            current_day = self.stop - datetime.timedelta(days=real_duration - 1)

            # check depends
            if self.depends_of is not None:
                prev_task_end = self.depends_of[0].end_date()
                for t in self.depends_of:
                    if isinstance(t, Milestone):
                        if t.end_date() > prev_task_end:
                            prev_task_end = t.end_date()
                    elif isinstance(t, Task):
                        if t.end_date() > prev_task_end:
                            prev_task_end = t.end_date()
                    # if t.end_date() > prev_task_end:
                    #     __LOG__.debug('*** latest one {0} which end on {1}'.format(t.name, t.end_date()))
                    #     prev_task_end = t.end_date()

                if prev_task_end > current_day:
                    start = prev_task_end + datetime.timedelta(days=1)
                    # return prev_task_end
                else:
                    start = current_day

                while start.weekday() in _not_worked_days() or start in VACATIONS:
                    start = start + datetime.timedelta(days=1)

                depend_start_date = start

                if depend_start_date > current_day:
                    __LOG__.error(
                        '** Due to dependencies, Task "{0}", could not be finished on time (should start as last on {1} but will start on {2})'.format(
                            self.fullname, current_day, depend_start_date
                        )
                    )
                    self._cache_start_date = depend_start_date
                else:
                    # should be first day of start...
                    self._cache_start_date = depend_start_date
            else:
                # should be first day of start...
                self._cache_start_date = current_day

        if self._cache_start_date != self.start:
            __LOG__.warning(
                '** starting date for task "{0}" is changed from {1} to {2}'.format(
                    self.fullname, self.start, self._cache_start_date
                )
            )
        return self._cache_start_date

    def set_end_date(self, end_date):
        """
        Set a end date, overriding the previous end date
        Parameters
        ----------
        end_date: datetime
            End date

        """

        self._cache_end_date = end_date
        self.end = end_date

    def end_date(self):
        """
        Returns the last day of the task, either the one which was given at task
        creation or the one calculated after checking dependencies
        """
        # Should take care of resources vacations ?
        if self._cache_end_date is not None:
            return self._cache_end_date

        __LOG__.debug("** Task::end_date ({0})".format(self.name))

        if self.duration is None or self.start is None and self.stop is not None:
            real_end = self.stop
            # Take care of vacations
            while real_end.weekday() in _not_worked_days() or real_end in VACATIONS:
                real_end -= datetime.timedelta(days=1)

            if real_end <= self.start_date():
                current_day = self.start_date()
                real_duration = 0
                if self.duration is None:
                    msg = (
                        f"End time {real_end} is before start time {self.start_date()} and no duration is given for\n"
                        f"project '{self.name}'. Please fix"
                    )
                    raise AssertionError(msg)
                duration: int = self.duration
                while duration > 1 or (
                    current_day.weekday() in _not_worked_days()
                    or current_day in VACATIONS
                ):
                    if not (
                        current_day.weekday() in _not_worked_days()
                        or current_day in VACATIONS
                    ):
                        real_duration = real_duration + 1
                        duration -= 1
                    else:
                        real_duration = real_duration + 1

                    current_day = self.start_date() + datetime.timedelta(
                        days=real_duration
                    )

                self._cache_end_date = self.start_date() + datetime.timedelta(
                    days=real_duration
                )
                __LOG__.warning(
                    '** task "{0}" will not be finished on time : end_date is changed from {1} to {2}'.format(
                        self.fullname, self.stop, self._cache_end_date
                    )
                )
                return self._cache_end_date

            self._cache_end_date = real_end
            if real_end != self.stop:
                __LOG__.warning(
                    '** task "{0}" will not be finished on time : end_date is changed from {1} to {2}'.format(
                        self.fullname, self.stop, self._cache_end_date
                    )
                )

            return self._cache_end_date

        if self.stop is None:
            current_day = self.start_date()
            real_duration = 0
            duration = self.duration
            while duration > 1 or (
                current_day.weekday() in _not_worked_days() or current_day in VACATIONS
            ):
                if not (
                    current_day.weekday() in _not_worked_days()
                    or current_day in VACATIONS
                ):
                    real_duration = real_duration + 1
                    duration -= 1
                else:
                    real_duration = real_duration + 1

                current_day = self.start_date() + datetime.timedelta(days=real_duration)

            self._cache_end_date = self.start_date() + datetime.timedelta(
                days=real_duration
            )
            return self._cache_end_date

        raise AssertionError("Something happend that should not")

    def svg(
        self,
        prev_y=0,
        start=None,
        end=None,
        planning_start=None,
        planning_end=None,
        color=None,
        level=None,
        scale=DRAW_WITH_DAILY_SCALE,
        title_align_on_left=False,
        offset=0,
    ):
        """
        Return SVG for drawing this task.

        Keyword arguments:
        prev_y -- int, line to start to draw
        start -- datetime.date of first day to draw
        end -- datetime.date of last day to draw
        color -- string of color for drawing the project
        level -- int, indentation level of the project, not used here
        scale -- drawing scale (d: days, w: weeks, m: months, q: quarterly)
        title_align_on_left -- boolean, align task title on left
        offset -- X offset from image border to start of drawing zone
        """
        __LOG__.debug(
            "** Task::svg ({0})".format(
                {
                    "name": self.name,
                    "prev_y": prev_y,
                    "start": start,
                    "end": end,
                    "color": color,
                    "level": level,
                }
            )
        )

        if not self.display:
            __LOG__.debug("** Task::svg ({0}) display off".format({"name": self.name}))
            return (None, 0)

        add_modified_begin_mark = False
        add_modified_end_mark = False

        if start is None:
            start = self.start_date()

        if self.start is not None and self.start_date() != self.start:
            add_modified_begin_mark = True

        if end is None:
            end = self.end_date()

        if self.stop is not None and self.end_date() != self.stop:
            add_modified_end_mark = True

        # override project color if defined
        if self.color is not None:
            color = self.color

        add_begin_mark = False
        add_end_mark = False

        y = prev_y * 10

        if scale == DRAW_WITH_DAILY_SCALE:

            def _time_diff(e, s):
                return (e - s).days

            def _time_diff_d(e, s):
                return _time_diff(e, s) + 1

        elif scale == DRAW_WITH_WEEKLY_SCALE:

            def _time_diff(end_date, start_date):
                td = 0
                guess = start_date
                while guess.weekday() != 0:
                    guess = guess + dateutil.relativedelta.relativedelta(days=-1)

                while end_date.weekday() != 6:
                    end_date = end_date + dateutil.relativedelta.relativedelta(days=+1)

                while guess + dateutil.relativedelta.relativedelta(days=+6) < end_date:
                    td += 1
                    guess = guess + dateutil.relativedelta.relativedelta(weeks=+1)

                return td

            def _time_diff_d(e, s):
                return _time_diff(e, s) + 1

        elif scale == DRAW_WITH_MONTHLY_SCALE:

            def _time_diff(end_date, start_date):
                return (
                    dateutil.relativedelta.relativedelta(end_date, start_date).months
                    + dateutil.relativedelta.relativedelta(end_date, start_date).years
                    * 12
                )

            def _time_diff_d(e, s):
                return _time_diff(e, s) + 1

        elif scale == DRAW_WITH_QUARTERLY_SCALE:
            raise ValueError("DRAW_WITH_QUARTERLY_SCALE not implemented yet")
        else:
            raise AssertionError(f"scale {scale} not recognised")

        # cas 1 -s--S==E--e-
        if self.start_date() >= start and self.end_date() <= end:
            x = _time_diff(self.start_date(), start) * 10
            d = _time_diff_d(self.end_date(), self.start_date()) * 10
            self.drawn_x_begin_coord = x
            self.drawn_x_end_coord = x + d
        # cas 5 -s--e--S==E-
        elif self.start_date() > end:
            return None, 0
        # cas 6 -S==E-s--e-
        elif self.end_date() < start:
            return None, 0
        # cas 2 -S==s==E--e-
        elif self.start_date() < start and self.end_date() <= end:
            x = 0
            d = _time_diff_d(self.end_date(), start) * 10
            self.drawn_x_begin_coord = x
            self.drawn_x_end_coord = x + d
            add_begin_mark = True
        # cas 3 -s--S==e==E-
        elif self.start_date() >= start and self.end_date() > end:
            x = _time_diff(self.start_date(), start) * 10
            d = _time_diff_d(end, self.start_date()) * 10
            self.drawn_x_begin_coord = x
            self.drawn_x_end_coord = x + d
            add_end_mark = True
        # cas 4 -S==s==e==E-
        elif self.start_date() < start and self.end_date() > end:
            x = 0
            d = _time_diff_d(end, start) * 10
            self.drawn_x_begin_coord = x
            self.drawn_x_end_coord = x + d
            add_end_mark = True
            add_begin_mark = True
        else:
            return None, 0

        self.drawn_y_coord = y

        svg = svgwrite.container.Group(id=re.sub(r"[ ,'/()]", "_", self.name))
        svg.add(
            svgwrite.shapes.Rect(
                insert=((x + 1 + offset) * mm, (y + 1) * mm),
                size=((d - 2) * mm, 8 * mm),
                fill=color,
                stroke=color,
                stroke_width=2,
                opacity=0.85,
            )
        )
        svg.add(
            svgwrite.shapes.Rect(
                insert=((x + 1 + offset) * mm, (y + 6) * mm),
                size=((d - 2) * mm, 3 * mm),
                fill="#909090",
                stroke=color,
                stroke_width=1,
                opacity=0.2,
            )
        )

        if add_modified_begin_mark:
            svg.add(
                svgwrite.shapes.Rect(
                    insert=((x + 1) * mm, (y + 1) * mm),
                    size=(5 * mm, 4 * mm),
                    fill="#0000FF",
                    stroke=color,
                    stroke_width=1,
                    opacity=0.35,
                )
            )

        if add_modified_end_mark:
            svg.add(
                svgwrite.shapes.Rect(
                    insert=((x + d - 7 + 1) * mm, (y + 1) * mm),
                    size=(5 * mm, 4 * mm),
                    fill="#0000FF",
                    stroke=color,
                    stroke_width=1,
                    opacity=0.35,
                )
            )

        if add_begin_mark:
            svg.add(
                svgwrite.shapes.Rect(
                    insert=((x + 1) * mm, (y + 1) * mm),
                    size=(5 * mm, 8 * mm),
                    fill="#000000",
                    stroke=color,
                    stroke_width=1,
                    opacity=0.2,
                )
            )
        if add_end_mark:
            svg.add(
                svgwrite.shapes.Rect(
                    insert=((x + d - 7 + 1) * mm, (y + 1) * mm),
                    size=(5 * mm, 8 * mm),
                    fill="#000000",
                    stroke=color,
                    stroke_width=1,
                    opacity=0.2,
                )
            )

        if self.percent_done is not None and self.percent_done > 0:
            # Bar shade
            svg.add(
                svgwrite.shapes.Rect(
                    insert=((x + 1 + offset) * mm, (y + 6) * mm),
                    size=(((d - 2) * self.percent_done / 100) * mm, 3 * mm),
                    fill="#F08000",
                    stroke=color,
                    stroke_width=1,
                    opacity=0.35,
                )
            )

        if not title_align_on_left:
            tx = x + 2
        else:
            tx = 5

        svg.add(
            svgwrite.text.Text(
                self.fullname,
                insert=(tx * mm, (y + 5) * mm),
                fill=_font_attributes()["fill"],
                stroke=_font_attributes()["stroke"],
                stroke_width=_font_attributes()["stroke_width"],
                font_family=_font_attributes()["font_family"],
                font_size=15,
            )
        )

        if self.resources is not None:
            t = " / ".join(["{0}".format(r.name) for r in self.resources])
            svg.add(
                svgwrite.text.Text(
                    "{0}".format(t),
                    insert=(tx * mm, (y + 8.5) * mm),
                    fill="purple",
                    stroke=_font_attributes()["stroke"],
                    stroke_width=_font_attributes()["stroke_width"],
                    font_family=_font_attributes()["font_family"],
                    font_size=15 - 5,
                )
            )

        return svg, 1

    def svg_dependencies(self, prj):
        """
        Draws svg dependencies between task and project according to coordinates
        cached when drawing tasks

        Keyword arguments:
        prj -- Project object to check against
        """
        __LOG__.debug(
            "** Task::svg_dependencies ({0})".format({"name": self.name, "prj": prj})
        )
        if self.depends_of is None:
            return None
        else:
            svg = svgwrite.container.Group()
            for t in self.depends_of:
                if isinstance(t, Milestone):
                    if not (
                        t.drawn_x_end_coord is None
                        or t.drawn_y_coord is None
                        or self.drawn_x_begin_coord is None
                    ) and prj.is_in_project(t):
                        if t.drawn_x_end_coord < self.drawn_x_begin_coord:
                            # horizontal line
                            svg.add(
                                svgwrite.shapes.Line(
                                    start=(
                                        (t.drawn_x_end_coord + 9) * mm,
                                        (t.drawn_y_coord + 5) * mm,
                                    ),
                                    end=(
                                        self.drawn_x_begin_coord * mm,
                                        (t.drawn_y_coord + 5) * mm,
                                    ),
                                    stroke="black",
                                    stroke_dasharray="5,3",
                                )
                            )

                            marker = svgwrite.container.Marker(
                                insert=(5, 5), size=(10, 10)
                            )
                            marker.add(
                                svgwrite.shapes.Circle(
                                    (5, 5),
                                    r=5,
                                    fill="#000000",
                                    opacity=0.5,
                                    stroke_width=0,
                                )
                            )
                            svg.add(marker)
                            # vertical line
                            eline = svgwrite.shapes.Line(
                                start=(
                                    self.drawn_x_begin_coord * mm,
                                    (t.drawn_y_coord + 5) * mm,
                                ),
                                end=(
                                    self.drawn_x_begin_coord * mm,
                                    (self.drawn_y_coord + 5) * mm,
                                ),
                                stroke="black",
                                stroke_dasharray="5,3",
                            )
                            eline["marker-end"] = marker.get_funciri()
                            svg.add(eline)

                        else:
                            # horizontal line
                            svg.add(
                                svgwrite.shapes.Line(
                                    start=(
                                        (t.drawn_x_end_coord + 9) * mm,
                                        (t.drawn_y_coord + 5) * mm,
                                    ),
                                    end=(
                                        (self.drawn_x_begin_coord + 10) * mm,
                                        (t.drawn_y_coord + 5) * mm,
                                    ),
                                    stroke="black",
                                    stroke_dasharray="5,3",
                                )
                            )
                            # vertical
                            svg.add(
                                svgwrite.shapes.Line(
                                    start=(
                                        (self.drawn_x_begin_coord + 10) * mm,
                                        (t.drawn_y_coord + 5) * mm,
                                    ),
                                    end=(
                                        (self.drawn_x_begin_coord + 10) * mm,
                                        (t.drawn_y_coord + 15) * mm,
                                    ),
                                    stroke="black",
                                    stroke_dasharray="5,3",
                                )
                            )
                            # horizontal line
                            svg.add(
                                svgwrite.shapes.Line(
                                    start=(
                                        self.drawn_x_begin_coord * mm,
                                        (t.drawn_y_coord + 15) * mm,
                                    ),
                                    end=(
                                        (self.drawn_x_begin_coord + 10) * mm,
                                        (t.drawn_y_coord + 15) * mm,
                                    ),
                                    stroke="black",
                                    stroke_dasharray="5,3",
                                )
                            )

                            marker = svgwrite.container.Marker(
                                insert=(5, 5), size=(10, 10)
                            )
                            marker.add(
                                svgwrite.shapes.Circle(
                                    (5, 5),
                                    r=5,
                                    fill="#000000",
                                    opacity=0.5,
                                    stroke_width=0,
                                )
                            )
                            svg.add(marker)
                            # vertical line
                            eline = svgwrite.shapes.Line(
                                start=(
                                    self.drawn_x_begin_coord * mm,
                                    (t.drawn_y_coord + 15) * mm,
                                ),
                                end=(
                                    self.drawn_x_begin_coord * mm,
                                    (self.drawn_y_coord + 5) * mm,
                                ),
                                stroke="black",
                                stroke_dasharray="5,3",
                            )
                            eline["marker-end"] = marker.get_funciri()
                            svg.add(eline)

                elif isinstance(t, Task):
                    if not (
                        t.drawn_x_end_coord is None
                        or t.drawn_y_coord is None
                        or self.drawn_x_begin_coord is None
                    ) and prj.is_in_project(t):
                        # horizontal line
                        svg.add(
                            svgwrite.shapes.Line(
                                start=(
                                    (t.drawn_x_end_coord - 2) * mm,
                                    (t.drawn_y_coord + 5) * mm,
                                ),
                                end=(
                                    self.drawn_x_begin_coord * mm,
                                    (t.drawn_y_coord + 5) * mm,
                                ),
                                stroke="black",
                                stroke_dasharray="5,3",
                            )
                        )

                        marker = svgwrite.container.Marker(insert=(5, 5), size=(10, 10))
                        marker.add(
                            svgwrite.shapes.Circle(
                                (5, 5), r=5, fill="#000000", opacity=0.5, stroke_width=0
                            )
                        )
                        svg.add(marker)
                        # vertical line
                        eline = svgwrite.shapes.Line(
                            start=(
                                self.drawn_x_begin_coord * mm,
                                (t.drawn_y_coord + 5) * mm,
                            ),
                            end=(
                                self.drawn_x_begin_coord * mm,
                                (self.drawn_y_coord + 5) * mm,
                            ),
                            stroke="black",
                            stroke_dasharray="5,3",
                        )
                        eline["marker-end"] = marker.get_funciri()
                        svg.add(eline)

        return svg

    def nb_elements(self):
        """
        Returns the number of task, 1 here
        """
        __LOG__.debug("** Task::nb_elements ({0})".format({"name": self.name}))
        return 1

    def _reset_coord(self):
        """
        Reset cached elements of task
        """
        __LOG__.debug("** Task::reset_coord ({0})".format({"name": self.name}))
        self.drawn_x_begin_coord = None
        self.drawn_x_end_coord = None
        self.drawn_y_coord = None
        self._cache_start_date = None
        self._cache_end_date = None
        return

    def is_in_project(self, task):
        """
        Return True if the given Task is itself... (lazy coding ;)

        Keyword arguments:
        task -- Task object
        """
        __LOG__.debug(
            "** Task::is_in_project ({0})".format({"name": self.name, "task": task})
        )
        if task is self:
            return True

        return False

    def get_resources(self):
        """
        Returns Resources used in the task
        """
        return self.resources

    def check_conflicts_between_task_and_resources_vacations(self):
        """
        Displays a warning for each conflict between tasks and vacation of
        resources affected to the task

        And returns a dictionary for resource vacation conflicts
        """
        conflicts = []
        if self.get_resources() is None:
            return conflicts
        for r in self.get_resources():
            cday = self.start_date()
            while cday <= self.end_date():
                if cday.weekday() not in _not_worked_days() and not r.is_available(
                    cday
                ):
                    conflicts.append(
                        {"resource": r.name, "date": cday, "task": self.name}
                    )
                    __LOG__.warning(
                        '** Caution resource "{0}" is affected on task "{2}" during vacations on day {1}'.format(
                            r.name, cday, self.fullname
                        )
                    )
                cday += datetime.timedelta(days=1)
        return conflicts

    def csv(self, csv=None):
        """
        Create CSV output from tasks

        Keyword arguments:
        csv -- None, dymmy object
        """
        if self.resources is not None:
            resources = ", ".join([x.fullname for x in self.resources])
        else:
            resources = ""

        csv_text = '"{0}";"{1}";{2};{3};{4};"{5}";\r\n'.format(
            self.state.replace('"', '\\"'),
            self.fullname.replace('"', '\\"'),
            self.start_date(),
            self.end_date(),
            self.duration,
            resources.replace('"', '\\"'),
        )
        return csv_text


############################################################################


class Milestone(Task):
    """
    Class for manipulating Milestones
    """

    def __init__(
        self, name, start=None, depends_of=None, color=None, fullname=None, display=True
    ):
        """
        Initialize milestone object. Two of start, stop or duration may be given.
        This milestone can rely on other milestone and will be completed with resources.
        If percent done is given, a progress bar will be included on the milestone.
        If color is specified, it will be used for the milestone.

        Keyword arguments:
        name -- name of the milestone (id)
        fullname -- long name given to the resource
        start -- datetime.date, first day of the milestone, default None
        depends_of -- list of Milestone which are parents of this one, default None
        color -- string, html color, default None
        display -- boolean, display this milestone, default True
        """
        __LOG__.debug(
            "** Milestone::__init__ {0}".format(
                {"name": name, "start": start, "depends_of": depends_of}
            )
        )
        self.name = name
        if fullname is not None:
            self.fullname = fullname
        else:
            self.fullname = name

        self.start = start
        self.stop = start
        self.duration = 0
        if color is not None:
            self.color = color
        else:
            self.color = "#FF3030"

        self.display = display
        self.state = "Milestone"

        if type(depends_of) is type([]):
            self.depends_of = depends_of
        elif depends_of is not None:
            self.depends_of = [depends_of]
        else:
            self.depends_of = None

        self.drawn_x_begin_coord = None
        self.drawn_x_end_coord = None
        self.drawn_y_coord = None
        self._cache_start_date = None
        self._cache_end_date = None

        return

    def end_date(self):
        """
        Returns the last day of the milestone, either the one which was given at milestone
        creation or the one calculated after checking dependencies
        """
        __LOG__.debug("** Milestone::end_date ({0})".format(self.name))
        # return self.start_date() - datetime.timedelta(days=1)
        return self.start_date()

    def svg(
        self,
        prev_y=0,
        start=None,
        end=None,
        planning_start=None,
        planning_end=None,
        color=None,
        level=None,
        scale=DRAW_WITH_DAILY_SCALE,
        title_align_on_left=False,
        offset=0,
    ):
        """
        Return SVG for drawing this milestone.

        Keyword arguments:
        prev_y -- int, line to start to draw
        start -- datetime.date of first day to draw
        end -- datetime.date of last day to draw
        planning_start -- datetime.date start date of planning
        planning_end -- datetime.date  end date of planning
        color -- string of color for drawing the project
        level -- int, indentation level of the project, not used here
        scale -- drawing scale (d: days, w: weeks, m: months, q: quarterly)
        title_align_on_left -- boolean, align milestone title on left
        offset -- X offset from image border to start of drawing zone
        """
        __LOG__.debug(
            "** Milestone::svg ({0})".format(
                {
                    "name": self.name,
                    "prev_y": prev_y,
                    "start": start,
                    "end": end,
                    "color": color,
                    "level": level,
                }
            )
        )

        if not self.display:
            __LOG__.debug(
                "** Milestone::svg ({0}) display off".format({"name": self.name})
            )
            return (None, 0)

        # add_modified_begin_mark = False
        # add_modified_end_mark = False

        if start is None:
            start = self.start_date()

        # if self.start_date() != self.start and self.start is not None:
        #    add_modified_begin_mark = True

        if end is None:
            end = self.end_date()

        # if self.end_date() != self.stop and self.stop is not None:
        #    add_modified_end_mark = True

        # override project color if defined
        if self.color is not None:
            color = self.color

        # add_begin_mark = False
        # add_end_mark = False

        y = prev_y * 10

        if scale == DRAW_WITH_DAILY_SCALE:

            def _time_diff(e, s):
                return (e - s).days

            def _time_diff_d(e, s):
                return _time_diff(e, s) + 1

        elif scale == DRAW_WITH_WEEKLY_SCALE:

            def _time_diff(end_date, start_date):
                td = 0
                guess = start_date
                # find first day of the week
                while guess.weekday() != 0:
                    guess = guess + dateutil.relativedelta.relativedelta(days=-1)
                # find last day of the week
                while end_date.weekday() != 6:
                    end_date = end_date + dateutil.relativedelta.relativedelta(days=+1)

                while guess <= end_date:
                    td += 1
                    guess = guess + dateutil.relativedelta.relativedelta(weeks=+1)

                return td - 1

            def _time_diff_d(e, s):
                return _time_diff(e, s) + 1

        elif scale == DRAW_WITH_MONTHLY_SCALE:

            def _time_diff(end_date, start_date):
                return (
                    dateutil.relativedelta.relativedelta(end_date, start_date).months
                    + dateutil.relativedelta.relativedelta(end_date, start_date).years
                    * 12
                )

            def _time_diff_d(e, s):
                return _time_diff(e, s) + 1

        elif scale == DRAW_WITH_QUARTERLY_SCALE:
            raise ValueError("DRAW_WITH_QUARTERLY_SCALE not implemented yet")
        else:
            raise AssertionError(f"scale {scale} not recognised")

        # cas 1 -s--X--e-
        if self.start_date() >= start and self.end_date() <= end:
            x = _time_diff(self.start_date(), start) * 10
            self.drawn_x_begin_coord = x
            self.drawn_x_end_coord = x
        else:
            return None, 0

        self.drawn_y_coord = y

        # insert=((x+1)*mm, (y+1)*mm),
        # size=((d-2)*mm, 8*mm),

        svg = svgwrite.container.Group(id=re.sub(r"[ ,'/()]", "_", self.name))
        # 3.543307 is for conversion from mm to pt units !
        svg.add(
            svgwrite.shapes.Polygon(
                points=[
                    ((x + 5 + offset) * mm, (y + 2) * mm),
                    ((x + 8 + offset) * mm, (y + 5) * mm),
                    ((x + 5 + offset) * mm, (y + 8) * mm),
                    ((x + 2 + offset) * mm, (y + 5) * mm),
                ],
                fill=color,
                stroke=color,
                stroke_width=2,
                opacity=0.85,
            )
        )

        if not title_align_on_left:
            tx = x + 2
        else:
            tx = 5

        svg.add(
            svgwrite.text.Text(
                self.fullname,
                insert=(tx * mm, (y + 5) * mm),
                fill=_font_attributes()["fill"],
                stroke=_font_attributes()["stroke"],
                stroke_width=_font_attributes()["stroke_width"],
                font_family=_font_attributes()["font_family"],
                font_size=15,
            )
        )

        return svg, 2

    def svg_dependencies(self, prj):
        """
        Draws svg dependencies between milestone and project according to coordinates
        cached when drawing milestones

        Keyword arguments:
        prj -- Project object to check against
        """
        __LOG__.debug(
            "** Milestone::svg_dependencies ({0})".format(
                {"name": self.name, "prj": prj}
            )
        )
        if self.depends_of is None:
            return None
        else:
            svg = svgwrite.container.Group()
            for t in self.depends_of:
                if isinstance(t, Milestone):
                    if not (
                        t.drawn_x_end_coord is None
                        or t.drawn_y_coord is None
                        or self.drawn_x_begin_coord is None
                    ) and prj.is_in_project(t):
                        # horizontal line
                        svg.add(
                            svgwrite.shapes.Line(
                                start=(
                                    (t.drawn_x_end_coord + 9) * mm,
                                    (t.drawn_y_coord + 5) * mm,
                                ),
                                end=(
                                    (self.drawn_x_begin_coord + 5) * mm,
                                    (t.drawn_y_coord + 5) * mm,
                                ),
                                stroke="black",
                                stroke_dasharray="5,3",
                            )
                        )

                        marker = svgwrite.container.Marker(insert=(5, 5), size=(10, 10))
                        marker.add(
                            svgwrite.shapes.Circle(
                                (5, 5), r=5, fill="#000000", opacity=0.5, stroke_width=0
                            )
                        )
                        svg.add(marker)
                        # vertical line
                        eline = svgwrite.shapes.Line(
                            start=(
                                (self.drawn_x_begin_coord + 5) * mm,
                                (t.drawn_y_coord + 5) * mm,
                            ),
                            end=(
                                (self.drawn_x_begin_coord + 5) * mm,
                                self.drawn_y_coord * mm,
                            ),
                            stroke="black",
                            stroke_dasharray="5,3",
                        )
                        eline["marker-end"] = marker.get_funciri()
                        svg.add(eline)

                elif isinstance(t, Task):
                    if not (
                        t.drawn_x_end_coord is None
                        or t.drawn_y_coord is None
                        or self.drawn_x_begin_coord is None
                    ) and prj.is_in_project(t):
                        # horizontal line
                        svg.add(
                            svgwrite.shapes.Line(
                                start=(
                                    (t.drawn_x_end_coord - 2) * mm,
                                    (t.drawn_y_coord + 5) * mm,
                                ),
                                end=(
                                    (self.drawn_x_begin_coord + 5) * mm,
                                    (t.drawn_y_coord + 5) * mm,
                                ),
                                stroke="black",
                                stroke_dasharray="5,3",
                            )
                        )

                        marker = svgwrite.container.Marker(insert=(5, 5), size=(10, 10))
                        marker.add(
                            svgwrite.shapes.Circle(
                                (5, 5), r=5, fill="#000000", opacity=0.5, stroke_width=0
                            )
                        )
                        svg.add(marker)
                        # vertical line
                        eline = svgwrite.shapes.Line(
                            start=(
                                (self.drawn_x_begin_coord + 5) * mm,
                                (t.drawn_y_coord + 5) * mm,
                            ),
                            end=(
                                (self.drawn_x_begin_coord + 5) * mm,
                                (self.drawn_y_coord + 0) * mm,
                            ),
                            stroke="black",
                            stroke_dasharray="5,3",
                        )
                        eline["marker-end"] = marker.get_funciri()
                        svg.add(eline)

        return svg

    def get_resources(self):
        """
        Returns Resources used in the milestone
        """
        return []

    def check_conflicts_between_task_and_resources_vacations(self):
        """
        Displays a warning for each conflict between milestones and vacation of
        resources affected to the milestone

        And returns a dictionary for resource vacation conflicts
        """
        return []

    def csv(self, csv=None):
        """
        Create CSV output from milestones

        Keyword arguments:
        csv -- None, dymmy object
        """
        if self.resources is not None:
            resources = ", ".join([x.fullname for x in self.resources])
        else:
            resources = ""

        csv_text = '"{0}";"{1}";{2};{3};{4};"{5}";\r\n'.format(
            self.state.replace('"', '\\"'),
            self.fullname.replace('"', '\\"'),
            self.start_date(),
            self.end_date(),
            self.duration,
            resources.replace('"', '\\"'),
        )
        return csv_text


class Project:
    """
    Class for handling projects
    """

    def __init__(
        self,
        name="",
        color=None,
        side_bar_color=None,
        project_start=None,
        project_end=None,
        font=None,
    ):
        """
        Initialize project with a given name and color for all tasks

        Keyword arguments:
        name -- string, name of the project
        color -- color for all tasks of the project
        """
        self.tasks = []
        self.name = name
        if color is None:
            self.color = "#FFFF90"
        else:
            self.color = color

        self.font = font

        self.project_start = project_start
        self.project_end = project_end

        if side_bar_color is None:
            self.side_bar_color = self.color
        else:
            self.side_bar_color = side_bar_color

        self.cache_nb_elements = None
        return

    def add_task(self, task):
        """
        Add a Task to the Project. Task can also be a subproject

        Keyword arguments:
        task -- Task or Project object
        """
        self.tasks.append(task)
        self.cache_nb_elements = None
        return

    @staticmethod
    def _svg_calendar(
        maxx, maxy, start_date, today=None, scale=DRAW_WITH_DAILY_SCALE, offset=0
    ):
        """
        Draw calendar in svg, beginning at start_date for maxx days, containing
        maxy lines. If today is given, draw a blue line at date

        Keyword arguments:
        maxx -- number of days, weeks, months or quarters (depending on scale) to draw
        maxy -- number of lines to draw
        start_date -- datetime.date of the first day to draw
        today -- datetime.date of day as today reference
        scale -- drawing scale (d: days, w: weeks, m: months, q: quarterly)
        offset -- X offset from image border to start of drawing zone
        """
        dwg = svgwrite.container.Group()

        cal = {0: "Mo", 1: "Tu", 2: "We", 3: "Th", 4: "Fr", 5: "Sa", 6: "Su"}

        maxx += 1

        vlines = dwg.add(svgwrite.container.Group(id="vlines", stroke="lightgray"))
        for x in range(maxx):
            vlines.add(
                svgwrite.shapes.Line(
                    start=((x + offset / 10) * cm, 2 * cm),
                    end=((x + offset / 10) * cm, (maxy + 2) * cm),
                )
            )
            if scale == DRAW_WITH_DAILY_SCALE:
                jour = start_date + datetime.timedelta(days=x)
            elif scale == DRAW_WITH_WEEKLY_SCALE:
                jour = start_date + dateutil.relativedelta.relativedelta(weeks=+x)
            elif scale == DRAW_WITH_MONTHLY_SCALE:
                jour = start_date + dateutil.relativedelta.relativedelta(months=+x)
            elif scale == DRAW_WITH_QUARTERLY_SCALE:
                raise ValueError("DRAW_WITH_QUARTERLY_SCALE not implemented yet")
            else:
                raise AssertionError(f"scale {scale} not recognised")

            if today is not None and today == jour:
                vlines.add(
                    svgwrite.shapes.Rect(
                        insert=((x + 0.4 + offset) * cm, 2 * cm),
                        size=(0.2 * cm, maxy * cm),
                        fill="#76e9ff",
                        stroke="lightgray",
                        stroke_width=0,
                        opacity=0.8,
                    )
                )

            if scale == DRAW_WITH_DAILY_SCALE:
                # draw vacations
                if (
                    start_date + datetime.timedelta(days=x)
                ).weekday() in _not_worked_days() or (
                    start_date + datetime.timedelta(days=x)
                ) in VACATIONS:
                    vlines.add(
                        svgwrite.shapes.Rect(
                            insert=((x + offset / 10) * cm, 2 * cm),
                            size=(1 * cm, maxy * cm),
                            fill="gray",
                            stroke="lightgray",
                            stroke_width=1,
                            opacity=0.7,
                        )
                    )

                # Current day
                vlines.add(
                    svgwrite.text.Text(
                        "{1} {0:02}".format(jour.day, cal[jour.weekday()][0]),
                        insert=((x * 10 + 1 + offset) * mm, 19 * mm),
                        fill="black",
                        stroke="black",
                        stroke_width=0,
                        font_family=_font_attributes()["font_family"],
                        font_size=15 - 3,
                    )
                )
                # Year
                if jour.day == 1 and jour.month == 1:
                    vlines.add(
                        svgwrite.text.Text(
                            "{0}".format(jour.year),
                            insert=((x * 10 + 1 + offset) * mm, 5 * mm),
                            fill="#400000",
                            stroke="#400000",
                            stroke_width=0,
                            font_family=_font_attributes()["font_family"],
                            font_size=15 + 5,
                            font_weight="bold",
                        )
                    )
                # Month name
                if jour.day == 1:
                    vlines.add(
                        svgwrite.text.Text(
                            "{0}".format(jour.strftime("%B")),
                            insert=((x * 10 + 1 + offset) * mm, 10 * mm),
                            fill="#800000",
                            stroke="#800000",
                            stroke_width=0,
                            font_family=_font_attributes()["font_family"],
                            font_size=15 + 3,
                            font_weight="bold",
                        )
                    )
                # Week number
                if jour.weekday() == 0:
                    vlines.add(
                        svgwrite.text.Text(
                            "{0:02}".format(jour.isocalendar()[1]),
                            insert=((x * 10 + 1 + offset) * mm, 15 * mm),
                            fill="black",
                            stroke="black",
                            stroke_width=0,
                            font_family=_font_attributes()["font_family"],
                            font_size=15 + 1,
                            font_weight="bold",
                        )
                    )

            elif scale == DRAW_WITH_WEEKLY_SCALE:
                # Year
                if jour.isocalendar()[1] == 1 and jour.month == 1:
                    vlines.add(
                        svgwrite.text.Text(
                            "{0}".format(jour.year),
                            insert=((x * 10 + 1 + offset) * mm, 5 * mm),
                            fill="#400000",
                            stroke="#400000",
                            stroke_width=0,
                            font_family=_font_attributes()["font_family"],
                            font_size=15 + 5,
                            font_weight="bold",
                        )
                    )
                # Month name
                if jour.day <= 7:
                    vlines.add(
                        svgwrite.text.Text(
                            "{0}".format(jour.strftime("%B")),
                            insert=((x * 10 + 1 + offset) * mm, 10 * mm),
                            fill="#800000",
                            stroke="#800000",
                            stroke_width=0,
                            font_family=_font_attributes()["font_family"],
                            font_size=15 + 3,
                            font_weight="bold",
                        )
                    )
                vlines.add(
                    svgwrite.text.Text(
                        "{0:02}".format(jour.isocalendar()[1]),
                        insert=((x * 10 + 1 + offset) * mm, 15 * mm),
                        fill="black",
                        stroke="black",
                        stroke_width=0,
                        font_family=_font_attributes()["font_family"],
                        font_size=15 + 1,
                        font_weight="bold",
                    )
                )

            elif scale == DRAW_WITH_MONTHLY_SCALE:
                # Month number
                vlines.add(
                    svgwrite.text.Text(
                        "{0}".format(jour.strftime("%m")),
                        insert=((x * 10 + 1 + offset) * mm, 19 * mm),
                        fill="black",
                        stroke="black",
                        stroke_width=0,
                        font_family=_font_attributes()["font_family"],
                        font_size=15 - 3,
                    )
                )
                # Year
                if jour.month == 1:
                    vlines.add(
                        svgwrite.text.Text(
                            "{0}".format(jour.year),
                            insert=((x * 10 + 1 + offset) * mm, 5 * mm),
                            fill="#400000",
                            stroke="#400000",
                            stroke_width=0,
                            font_family=_font_attributes()["font_family"],
                            font_size=15 + 5,
                            font_weight="bold",
                        )
                    )

            elif scale == DRAW_WITH_QUARTERLY_SCALE:
                raise ValueError("DRAW_WITH_QUARTERLY_SCALE not implemented yet")
            else:
                raise AssertionError(f"scale {scale} not recognised")

        vlines.add(
            svgwrite.shapes.Line(
                start=((maxx + offset / 10) * cm, 2 * cm),
                end=((maxx + offset / 10) * cm, (maxy + 2) * cm),
            )
        )

        hlines = dwg.add(svgwrite.container.Group(id="hlines", stroke="lightgray"))

        dwg.add(
            svgwrite.shapes.Line(
                start=((0 + offset / 10) * cm, (2) * cm),
                end=((maxx + offset / 10) * cm, (2) * cm),
                stroke="black",
            )
        )
        dwg.add(
            svgwrite.shapes.Line(
                start=((0 + offset / 10) * cm, (maxy + 2) * cm),
                end=((maxx + offset / 10) * cm, (maxy + 2) * cm),
                stroke="black",
            )
        )

        for y in range(2, maxy + 3):
            hlines.add(
                svgwrite.shapes.Line(
                    start=((0 + offset / 10) * cm, y * cm),
                    end=((maxx + offset / 10) * cm, y * cm),
                )
            )

        return dwg

    def make_svg_for_tasks(
        self,
        filename,
        today=None,
        start=None,
        end=None,
        margin_left=None,
        margin_right=None,
        scale=DRAW_WITH_DAILY_SCALE,
        title_align_on_left=False,
        offset=0,
    ):
        """
        Draw gantt of tasks and output it to filename. If start or end are
        given, use them as reference, otherwise use project first and last day

        Keyword arguments:
        filename -- string, filename to save to OR file object
        today -- datetime.date of day marked as a reference
        start -- datetime.date of first day to draw
        end -- datetime.date of last day to draw
        margin_left -- number of week to add to the grid before the project start
        margin_right -- number of week to add to the grid after the project end
        planning_start -- start of the planning
        planning_end -- end of the planning
        scale -- drawing scale (d: days, w: weeks, m: months, q: quarterly)
        title_align_on_left -- boolean, align task title on left
        offset -- X offset from image border to start of drawing zone
        """
        if len(self.tasks) == 0:
            __LOG__.warning("** Empty project : {0}".format(self.name))
            return

        self._reset_coord()

        if start is None:
            start_date = self.start_date()
        else:
            start_date = start

        if end is None:
            end_date = self.end_date()
        else:
            end_date = end

        # add a margin to the left and right of the right in order to make some space for the project labels
        if margin_left is not None:
            start_date -= datetime.timedelta(weeks=margin_left)
        if margin_right is not None:
            end_date += datetime.timedelta(weeks=margin_right)

        if start_date > end_date:
            __LOG__.critical(
                "start date {0} > end_date {1}".format(start_date, end_date)
            )
            sys.exit(1)

        ldwg = svgwrite.container.Group()
        psvg, pheight = self.svg(
            prev_y=2,
            start=start_date,
            end=end_date,
            planning_start=start,
            planning_end=end,
            color=self.side_bar_color,
            scale=scale,
            title_align_on_left=title_align_on_left,
            offset=offset,
        )
        if psvg is not None:
            ldwg.add(psvg)

        dep = self.svg_dependencies(self)
        if dep is not None:
            ldwg.add(dep)

        if scale == DRAW_WITH_DAILY_SCALE:
            # how many days do we need to draw ?
            maxx = (end_date - start_date).days
        elif scale == DRAW_WITH_WEEKLY_SCALE:
            # how many weeks do we need to draw ?
            maxx = 0

            guess = start_date
            while guess.weekday() != 0:
                guess = guess + dateutil.relativedelta.relativedelta(days=-1)

            while end_date.weekday() != 6:
                end_date = end_date + dateutil.relativedelta.relativedelta(days=+1)

            while guess <= end_date:
                maxx += 1
                guess = guess + dateutil.relativedelta.relativedelta(weeks=+1)
        elif scale == DRAW_WITH_MONTHLY_SCALE:
            # how many months do we need to draw ?
            if dateutil.relativedelta.relativedelta(end_date, start_date).days == 0:
                maxx = (
                    dateutil.relativedelta.relativedelta(end_date, start_date).months
                    + dateutil.relativedelta.relativedelta(end_date, start_date).years
                    * 12
                )
            else:
                maxx = (
                    dateutil.relativedelta.relativedelta(end_date, start_date).months
                    + dateutil.relativedelta.relativedelta(end_date, start_date).years
                    * 12
                    + 1
                )
        elif scale == DRAW_WITH_QUARTERLY_SCALE:
            raise ValueError("DRAW_WITH_QUARTERLY_SCALE not implemented yet")
        else:
            raise AssertionError(f"scale {scale} not recognised")

        dwg = _my_svgwrite_drawing_wrapper(filename, debug=True)
        dwg.add(
            svgwrite.shapes.Rect(
                insert=(0 * cm, 0 * cm),
                size=((maxx + 1 + offset / 10) * cm, (pheight + 3) * cm),
                fill="white",
                stroke_width=0,
                opacity=1,
            )
        )

        dwg.add(
            self._svg_calendar(maxx, pheight, start_date, today, scale, offset=offset)
        )
        dwg.add(ldwg)
        dwg.save(width=(maxx + 1 + offset / 10) * cm, height=(pheight + 3) * cm)
        return

    def make_svg_for_resources(
        self,
        filename,
        today=None,
        start=None,
        end=None,
        resources=None,
        one_line_for_tasks=False,
        tag_filter="",
        scale=DRAW_WITH_DAILY_SCALE,
        title_align_on_left=False,
        offset=0,
        color_per_taks=False,
    ):
        """
        Draw resources affectation and output it to filename. If start or end are
        given, use them as reference, otherwise use project first and last day

        And returns to a dictionary of dictionaries for vacation and task
        conflicts for resources

        Keyword arguments:
        filename -- string, filename to save to OR file object
        today -- datetime.date of day marked as a reference
        start -- datetime.date of first day to draw
        end -- datetime.date of last day to draw
        resources -- list of Resource to check, default all
        one_line_for_tasks -- use only one line to display all tasks ?
        tag_filter -- display only those tags
        scale -- drawing scale (d: days, w: weeks, m: months, q: quarterly)
        title_align_on_left -- boolean, align task title on left
        offset -- X offset from image border to start of drawing zone
        color_per_task -- give a color per task
        """

        if scale != DRAW_WITH_DAILY_SCALE:
            __LOG__.warning(
                "** Will draw ressource graph at day scale, not {0} as requested".format(
                    scale
                )
            )
            scale = DRAW_WITH_DAILY_SCALE

        if len(self.tasks) == 0:
            __LOG__.warning("** Empty project : {0}".format(self.name))
            return

        self._reset_coord()

        if start is None:
            start_date = self.start_date()
        else:
            start_date = start

        if end is None:
            end_date = self.end_date()
        else:
            end_date = end

        if start_date > end_date:
            __LOG__.critical(
                "start date {0} > end_date {1}".format(start_date, end_date)
            )
            sys.exit(1)

        if resources is None:
            resources = self.get_resources()

        maxx = (end_date - start_date).days
        maxy = len(resources) * 2

        if maxy == 0:
            # No resources
            return {}

        # detect conflicts between resources and holidays
        conflicts_vacations = []
        for t in self.get_tasks():
            conflicts_vacations.append(
                t.check_conflicts_between_task_and_resources_vacations()
            )

        conflicts_vacations = _flatten(conflicts_vacations)

        ldwg = svgwrite.container.Group()

        if not one_line_for_tasks:
            ldwg.add(
                svgwrite.shapes.Line(
                    start=((0) * cm, (2) * cm),
                    end=((maxx + 1 + offset / 10) * cm, (2) * cm),
                    stroke="black",
                )
            )

        nline = 2
        conflicts_tasks = []
        conflict_display_line = 1
        for r in resources:
            # do stuff for each resource
            if tag_filter != "" and r.name not in tag_filter:
                continue

            ress = svgwrite.container.Group()
            try:
                res_text = svgwrite.text.Text(
                    "{0}".format(r.fullname),
                    insert=(3 * mm, (nline * 10 + 7) * mm),
                    fill=_font_attributes()["fill"],
                    stroke=_font_attributes()["stroke"],
                    stroke_width=_font_attributes()["stroke_width"],
                    font_family=_font_attributes()["font_family"],
                    font_size=15 + 3,
                )
            except ValueError as err:
                __LOG__.warning(f"Failed making text object for {r.fullname}")
            else:
                ress.add(res_text)

            overcharged_days = r.search_for_task_conflicts()

            conflict_display_line = nline
            nline += 1

            vac = svgwrite.container.Group()
            conflicts = svgwrite.container.Group()
            cday = start_date
            while cday <= end_date:
                # Vacations
                if (
                    cday.weekday() not in _not_worked_days()
                    and cday not in VACATIONS
                    and not r.is_available(cday)
                ):
                    vac.add(
                        svgwrite.shapes.Rect(
                            insert=(
                                ((cday - start_date).days * 10 + 1 + offset) * mm,
                                (conflict_display_line * 10 + 1) * mm,
                            ),
                            size=(4 * mm, 8 * mm),
                            fill=COLOR_VACATION_DEFAULT,
                            stroke=COLOR_VACATION_DEFAULT,
                            stroke_width=1,
                            opacity=0.65,
                        )
                    )

                # Overcharge
                if (
                    cday.weekday() not in _not_worked_days()
                    and cday not in VACATIONS
                    and cday in overcharged_days
                ):
                    conflicts.add(
                        svgwrite.shapes.Rect(
                            insert=(
                                ((cday - start_date).days * 10 + 1 + 4 + offset) * mm,
                                (conflict_display_line * 10 + 1) * mm,
                            ),
                            size=(4 * mm, 8 * mm),
                            fill=COLOR_OVERCHARGE_DEFAULT,
                            stroke=COLOR_OVERCHARGE_DEFAULT,
                            stroke_width=1,
                            opacity=0.65,
                        )
                    )

                cday += datetime.timedelta(days=1)

            nb_tasks = 0
            for t in self.get_tasks():
                if t.get_resources() is not None and r in t.get_resources():
                    if color_per_taks:
                        color = t.color
                    else:
                        color = r.color

                    if color is None:
                        color = COLOR_RESOURCE_DEFAULT

                    psvg, void = t.svg(
                        prev_y=nline,
                        start=start_date,
                        end=end_date,
                        color=color,
                        scale=scale,
                        title_align_on_left=title_align_on_left,
                        offset=offset,
                    )
                    if psvg is not None:
                        ldwg.add(psvg)
                        nb_tasks += 1
                        if not one_line_for_tasks:
                            nline += 1

            if nb_tasks == 0:
                nline -= 1
            elif nb_tasks > 0:
                __LOG__.info(r.fullname, nb_tasks)
                ldwg.add(ress)
                ldwg.add(vac)
                ldwg.add(conflicts)

                if not one_line_for_tasks:
                    ldwg.add(
                        svgwrite.shapes.Line(
                            start=(0 * cm, nline * cm),
                            end=((maxx + 1 + offset / 10) * cm, nline * cm),
                            stroke="black",
                        )
                    )

                # nline += 1
                if one_line_for_tasks:
                    nline += 1
                    ldwg.add(
                        svgwrite.shapes.Line(
                            start=(0 * cm, nline * cm),
                            end=((maxx + 1) * cm, nline * cm),
                            stroke="black",
                        )
                    )

        dwg = _my_svgwrite_drawing_wrapper(filename, debug=True)
        dwg.add(
            svgwrite.shapes.Rect(
                insert=(0 * cm, 0 * cm),
                size=((maxx + 1 + offset / 10) * cm, (nline + 1) * cm),
                fill="white",
                stroke_width=0,
                opacity=1,
            )
        )
        dwg.add(
            self._svg_calendar(maxx, nline - 2, start_date, today, scale, offset=offset)
        )
        dwg.add(ldwg)
        dwg.save(width=(maxx + 1 + offset / 10) * cm, height=(nline + 1) * cm)
        return {
            "conflicts_vacations": conflicts_vacations,
            "conflicts_tasks": conflicts_tasks,
        }

    def start_date(self):
        """
        Returns first day of the project
        """
        if len(self.tasks) == 0:
            __LOG__.warning("** Empty project : {0}".format(self.name))
            return datetime.date(9999, 1, 1)

        first = self.tasks[0].start_date()
        for t in self.tasks:
            if t.start_date() < first:
                first = t.start_date()
        return first

    def end_date(self):
        """
        Returns last day of the project
        """
        if len(self.tasks) == 0:
            __LOG__.warning("** Empty project : {0}".format(self.name))
            return datetime.date(1970, 1, 1)

        last = self.tasks[0].end_date()
        for t in self.tasks:
            if t.end_date() > last:
                last = t.end_date()
        return last

    def svg(
        self,
        prev_y=0,
        start=None,
        end=None,
        planning_start=None,
        planning_end=None,
        color=None,
        level=0,
        scale=DRAW_WITH_DAILY_SCALE,
        title_align_on_left=False,
        offset=0,
    ):
        """
        Return (SVG code, number of lines drawn) for the project. Draws all
        tasks and add project name with a purple bar on the left side.

        Keyword arguments:
        prev_y -- int, line to start to draw
        start -- datetime.date of first day to draw
        end -- datetime.date of last day to draw
        color -- string of color for drawing the project
        level -- int, indentation level of the project
        scale -- drawing scale (d: days, w: weeks, m: months, q: quarterly)
        title_align_on_left -- boolean, align task title on left
        offset -- X offset from image border to start of drawing zone
        """
        if start is None:
            start = self.start_date()
        if end is None:
            end = self.end_date()
        if color is None or self.color is not None:
            color = self.color

        cy = prev_y + 1 * (self.name != "")

        prj = svgwrite.container.Group()

        for t in self.tasks:
            if isinstance(t, Task):
                if not task_within_range(t, planning_start, planning_end):
                    continue

            trepr, theight = t.svg(
                cy,
                start=start,
                end=end,
                planning_start=planning_start,
                planning_end=planning_end,
                color=color,
                level=level + 1,
                scale=scale,
                title_align_on_left=title_align_on_left,
                offset=offset,
            )
            if trepr is not None:
                prj.add(trepr)
                cy += theight

        fprj = svgwrite.container.Group()
        prj_bar = False
        # if margin_left is not None:
        #     start += datetime.timedelta(days=margin_left)
        # if margin_right is not None:
        #     end -= datetime.timedelta(days=margin_right)
        if self.name != "":
            if (
                (self.start_date() >= start and self.end_date() <= end)
                or (self.end_date() >= start and self.start_date() <= end)
            ) or level == 1:
                # Adjust font level for level 0 (main project) and 1 (projects per project leader)
                # todo: allow modification in settings file
                if level == 0:
                    font_weight = "bold"
                    font_size = 18
                elif level == 1:
                    font_weight = "bold"
                    font_size = 16
                else:
                    font_weight = "normal"
                    font_size = 13
                fprj.add(
                    svgwrite.text.Text(
                        "{0}".format(self.name),
                        insert=(
                            (6 * level + 3 + offset) * mm,
                            ((prev_y) * 10 + 7) * mm,
                        ),
                        fill=_font_attributes()["fill"],
                        stroke=_font_attributes()["stroke"],
                        stroke_width=_font_attributes()["stroke_width"],
                        font_family=_font_attributes()["font_family"],
                        font_weight=font_weight,
                        font_size=font_size,
                    )
                )

                fprj.add(
                    svgwrite.shapes.Rect(
                        insert=((6 * level + 0.8 + offset) * mm, (prev_y + 0.5) * cm),
                        size=(0.2 * cm, ((cy - prev_y - 1) + 0.4) * cm),
                        fill=self.color,
                        stroke="lightgray",
                        stroke_width=0,
                        opacity=0.5,
                    )
                )
                prj_bar = True
            else:
                cy -= 1

        # Do not display empty tasks
        if (cy - prev_y) == 0 or ((cy - prev_y) == 1 and prj_bar):
            return None, 0

        fprj.add(prj)

        return fprj, cy - prev_y

    def svg_dependencies(self, prj):
        """
        Draws svg dependencies between tasks according to coordinates cached
        when drawing tasks

        Keyword arguments:
        prj -- Project object to check against
        """
        svg = svgwrite.container.Group()
        for t in self.tasks:
            trepr = t.svg_dependencies(prj)
            if trepr is not None:
                svg.add(trepr)
        return svg

    def nb_elements(self):
        """
        Returns the number of tasks included in the project or subproject
        """
        if self.cache_nb_elements is not None:
            return self.cache_nb_elements

        nb = 0
        for t in self.tasks:
            nb += t.nb_elements()

        self.cache_nb_elements = nb
        return nb

    def _reset_coord(self):
        """
        Reset cached elements of all tasks and project
        """
        self.cache_nb_elements = None
        for t in self.tasks:
            t._reset_coord()
        return

    def is_in_project(self, task):
        """
        Return True if the given Task is in the project, False if not

        Keyword arguments:
        task -- Task object
        """
        for t in self.tasks:
            if t.is_in_project(task):
                return True
        return False

    def get_resources(self):
        """
        Returns Resources used in the project
        """
        rlist = []
        for t in self.tasks:
            r = t.get_resources()
            if r is not None:
                rlist.append(r)

        flist = []
        for r in _flatten(rlist):
            if r not in flist:
                flist.append(r)
        return flist

    def get_tasks(self):
        """
        Returns flat list of Tasks used in the Project and subproject
        """
        tlist = []
        for t in self.tasks:
            # if it is a subproject, recurse
            if type(t) is type(self):
                st = t.get_tasks()
                tlist.append(st)
            else:  # get task
                tlist.append(t)

        flist = []
        for r in _flatten(tlist):
            if r not in flist:
                flist.append(r)
        return flist

    def csv(self, csv=None):
        """
        Create CSV output from projects

        Keyword arguments:
        csv -- string, filename to save to OR file object OR None
        """
        if len(self.tasks) == 0:
            __LOG__.warning("** Empty project : {0}".format(self.name))
            return

        if csv is not None:
            csv_text = bytes.decode(codecs.BOM_UTF8, "utf-8")
            csv_text += '"State";"Task Name";"Start date";"End date";"Duration";"Resources";\r\n'
        else:
            csv_text = ""

        for t in self.tasks:
            c = t.csv()
            if c is not None:
                csv_text += c

        if csv is not None:
            test = type(csv) == io.TextIOWrapper

            if test:
                csv.write(csv_text)
            else:
                with io.open(csv, mode="w", encoding="utf-8") as stream:
                    stream.write(csv_text)

        return csv_text


def task_within_range(task, planning_start, planning_end):
    """
    Check if task is in de range of the planning

    Parameters
    ----------
    task: object
        The task
    planning_start: datetime
        Start of the planning
    planning_end: datetime
        End of the planning

    Returns
    -------
    bool:
        True if task is in range
    """
    is_in_range = True
    if planning_start:
        try:
            task_start = task.start
        except AttributeError:
            pass
        else:
            if task_start < planning_start:
                is_in_range = False
    if planning_end:
        try:
            task_end = task.end
        except AttributeError:
            pass
        else:
            if task_end > planning_end:
                is_in_range = False
    return is_in_range


# MAIN -------------------
if __name__ == "__main__":
    import doctest

    # non regression test
    doctest.testmod()

else:
    init_log_to_sysout(level=logging.CRITICAL)


# <EOF>######################################################################



"""
Definition of CBS rbg colors. Based on the color rgb definitions from the cbs LaTeX template
"""

import logging

import webcolors

_logger = logging.getLogger(__name__)


# A bit tricky, but we use it to store the custom colors
CUSTOM_COLORS_HEX = {}


def hex_number_to_hex_hash(hex_number):
    """
    Args:
        hex_number: int or str
            Hexadecimal representation of a color, with or without a leading #

    Returns: str
        Hexadecimal color with a leading #
    """
    hex_code = str(hex_number)
    if set(list(hex_code.upper())).difference(set(list("1234567890ABCDEF#"))):
        raise ValueError(
            f"Hex color {hex_code} is not valid as it contains characters out of the valid range"
        )
    if not hex_code.startswith("#"):
        if "#" in hex_code:
            raise ValueError(
                f"Hex color {hex_code} is not valid as the # is not at the start"
            )
        hex_code = f"#{hex_code}"
    if len(hex_code) != 7:
        raise ValueError(
            f"Hex color {hex_code} is not valid as is has an invalid amount of digit "
        )
    return hex_code


def set_custom_colors(custom_colors):
    """Set the hex colors defined in the dictionary to the global variable CUSTOM_COLORS_HEX"""
    for color_name, color_hex in custom_colors.items():
        CUSTOM_COLORS_HEX[color_name] = hex_number_to_hex_hash(color_hex)


def color_to_hex(color: str):
    """Convert a named color into a hex code with a leading #"""
    color_hex_code = None
    if color is not None:
        try:
            color_hex_code = CUSTOM_COLORS_HEX[color]
            _logger.debug(f"color {color} met cbs colors omgezet naar {color_hex_code}")
        except KeyError:
            try:
                color_hex_code = webcolors.name_to_hex(color)
                _logger.debug(
                    f"color {color} met webcolors omgezet naar {color_hex_code}"
                )
            except (AttributeError, ValueError) as err:
                color_hex_code = hex_number_to_hex_hash(color)

    return color_hex_code



import pytest
import dateutil.parser as dparse

from gantt_project_maker.project_classes import get_nearest_saturday, parse_date

__author__ = "Eelco van Vliet"
__copyright__ = "Eelco van Vliet"
__license__ = "MIT"


def test_parse_date():
    date = "25-12-2023"
    assert parse_date(date, dayfirst=True) == dparse.parse(date, dayfirst=True).date()
    assert (
        parse_date(None, date_default=date, dayfirst=True)
        == dparse.parse(date, dayfirst=True).date()
    )


def test_nearest_saturday():
    """API Tests if check succeeds"""

    date = parse_date("25-12-2023", dayfirst=True)
    assert get_nearest_saturday(date) == parse_date("23-12-2023", dayfirst=True)

    date = parse_date("28-12-2023", dayfirst=True)
    assert get_nearest_saturday(date) == parse_date("30-12-2023", dayfirst=True)
